from puppeter import container
from puppeter.container import Named
from puppeter.domain.model.configurer import Configurer
from puppeter.domain.gateway.installer import InstallerGateway
from puppeter.persistence.gateway.installer.debian.pc3x import DebianPC3xConfigurer
from puppeter.persistence.gateway.installer.debian.pc4x import DebianPC4xConfigurer
from puppeter.persistence.gateway.installer.debian.pc5x import DebianPC5xConfigurer
from puppeter.persistence.gateway.installer.debian.system import DebianSystemConfigurer
from puppeter.persistence.gateway.installer.linux import LinuxInstallerGateway


@Named('debian')
class DebianInstallerGateway(LinuxInstallerGateway):

    def _provide_install_configurers(self, installer):
        return tuple([self.__provide_install_configurer(installer)])

    @staticmethod
    def __provide_install_configurer(installer):
        name = installer.bean_name()
        if name == 'gem':
            return container.get_named(Configurer, 'gem', installer=installer)
        name += '-debian'
        return container.get_named(Configurer, name, installer=installer)


container.bind(InstallerGateway, DebianInstallerGateway)
container.bind(Configurer, DebianPC3xConfigurer)
container.bind(Configurer, DebianPC4xConfigurer)
container.bind(Configurer, DebianPC5xConfigurer)
container.bind(Configurer, DebianSystemConfigurer)



from puppeter.container import Named
from puppeter.presentation.app import App


@Named('interactive')
class InteractiveApp(App):

    def __init__(self, options):
        App.__init__(self, options)

    def _collect_answers(self):
        raise NotImplementedError('Not yet implemented!')



from __future__ import absolute_import
import os
import distro
import platform
from typing import Sequence, Callable

import puppeter.settings
from puppeter.domain.facter import Facter
from puppeter.domain.model.osfacts import OsFamily, OperatingSystem, \
    OperatingSystemRelease, OperatingSystemCodename, Docker


def calculate_operatingsystem():
    if platform.system() == 'Linux':
        dist = distro.id()
        return {
            'fedora': OperatingSystem.Fedora,
            'centos': OperatingSystem.CentOS,
            'oracle': OperatingSystem.OracleLinux,
            'ol': OperatingSystem.OracleLinux,
            'rhel': OperatingSystem.RedHat,
            'scientific': OperatingSystem.Scientific,
            'debian': OperatingSystem.Debian,
            'ubuntu': OperatingSystem.Ubuntu,
            'sles': OperatingSystem.SLES,
            'opensuse': OperatingSystem.OpenSuse,
            'arch': OperatingSystem.ArchLinux
        }.get(dist, OperatingSystem.Unknown)
    return OperatingSystem.Unknown


def calculate_osfamily():
    return {
        OperatingSystem.CentOS: OsFamily.RedHat,
        OperatingSystem.OracleLinux: OsFamily.RedHat,
        OperatingSystem.RedHat: OsFamily.RedHat,
        OperatingSystem.Scientific: OsFamily.RedHat,
        OperatingSystem.Debian: OsFamily.Debian,
        OperatingSystem.Ubuntu: OsFamily.Debian,
        OperatingSystem.OpenSuse: OsFamily.Suse,
        OperatingSystem.SLES: OsFamily.Suse,
        OperatingSystem.ArchLinux: OsFamily.Arch
    }.get(Facter.get(OperatingSystem), OsFamily.Unknown)


def calculate_osrelease():
    if platform.system() == 'Linux':
        release = distro.linux_distribution(full_distribution_name=False)[1]
        return OperatingSystemRelease(release.strip())
    return None


def calculate_oscodename():
    if platform.system() == 'Linux':
        codename = distro.linux_distribution(full_distribution_name=False)[2]
        return OperatingSystemCodename(codename.strip().split(' ')[0].lower())
    return None


def calculate_docker():
    return Docker.YES if __calculate_docker_bool() else Docker.NO


def __calculate_docker_bool():
    if platform.system() != 'Linux':
        return False
    path = '{chroot}proc/1/cgroup'.format(chroot=puppeter.settings.chroot)
    if not __is_readable(path):
        return False
    try:
        lines = open(path).read().splitlines()
        return __any(lines, lambda l: '/docker' in l.split(':')[2])
    except IOError:
        return False


def __is_readable(file):
    return os.path.isfile(file) and os.access(file, os.R_OK)


def __any(seq, predicate):
    # type: (Sequence, Callable[[str], bool]) -> bool
    return any((predicate(i) is True) for i in seq)



import pytest
from puppeter.domain.model.gemrequirement import GemVersion, GemRequirement


def test_gem_version_release():
    # given
    v = GemVersion('1.2.4.beta')

    # when
    released = v.release()

    # then
    assert GemVersion('1.2.4') == released


def test_gem_version_bump():
    # given
    v = GemVersion('1.2.4')

    # when
    bumped = v.bump()

    # then
    assert GemVersion('1.3.0') == bumped


def test_gem_version_compare():
    assert GemVersion('1.3.0') == GemVersion('1.3')
    assert GemVersion('1.3.0') <= GemVersion('1.3')
    assert GemVersion('1.1.3') <= GemVersion('1.3')
    assert GemVersion('1.4.pre') >= GemVersion('1.3')
    assert GemVersion('1.4.pre') != GemVersion('1.4')


@pytest.mark.parametrize('requirement,version', [
    (['3.4'], '3.4.0'),
    (['~> 3.4'], '3.4.8'),
    (['>= 3.4'], '4.4.8'),
    (['>= 3.4', '<4'], '3.45.8')
])
def test_gem_requirement(requirement, version):
    assert GemRequirement(*requirement).satified_by(version)


@pytest.mark.parametrize('requirement,version', [
    (['>= 3.4', '<4'], '4.1'),
    (['~> 3'], '4.1.0.pre')
])
def test_gem_requirement_fails(requirement, version):
    assert GemRequirement(*requirement).satified_by(version) is False



from puppeter.container import Named
from puppeter.domain.facter import Facter
from puppeter.domain.model.ordered import Order
from puppeter.domain.model.osfacts import OperatingSystemRelease
from puppeter.persistence.gateway.installer.redhat.base import BaseRedHatConfigurer


@Named('system-redhat')
@Order(100)
class RedHatSystemConfigurer(BaseRedHatConfigurer):
    def _collect_repo(self, collector):
        rel = Facter.get(OperatingSystemRelease)
        collector.collect_from_file('Ensure EPEL is active', 'ensure-epel-{rel}.sh'.format(rel=rel.major()))

    def _collect_agent(self, collector):
        collector.collect_from_file('Puppet agent setup', 'puppet-system.sh')

    def _collect_server(self, collector):
        collector.collect_from_file('Puppet Master setup', 'puppetmaster-system.sh')



from __future__ import print_function

import abc

import attr
import re
import subprocess
import time
import timeit
import paramiko
import pytest
import os
import select
import colorama
import sys

import six
from paramiko import SFTPClient, SSHClient, Channel
from paramiko.common import o777
from typing import Sequence


class PhialLogger:
    PHIAL_WRAP = colorama.Fore.WHITE + colorama.Style.DIM + "phial >>> " + colorama.Style.RESET_ALL
    PHIAL_INFO = colorama.Fore.CYAN + "phial INFO: "

    def info(self, message, *args):
        self.__print("{phial}{line}{rs}".format(
            phial=PhialLogger.PHIAL_INFO,
            line=message % args,
            rs=colorama.Style.RESET_ALL
        ))

    def stdout(self, line):
        self.__print("{phial}{color}{line}{rs}".format(
            phial=PhialLogger.PHIAL_WRAP,
            color=colorama.Fore.BLUE,
            line=line,
            rs=colorama.Style.RESET_ALL
        ))

    def stderr(self, line):
        self.__print("{phial}{color}{line}{rs}".format(
            phial=PhialLogger.PHIAL_WRAP,
            color=colorama.Fore.RED,
            line=line,
            rs=colorama.Style.RESET_ALL
        ), file=sys.stderr)

    def __print(self, message, **kwargs):
        print(message, **kwargs)


class DirSFTPClient(paramiko.SFTPClient):
    def put_dir(self, source, target, exclude=('.git', '.tox')):
        """ Uploads the contents of the source directory to the target path. The
            target directory needs to exists. All subdirectories in source are
            created under target.
        """
        self.mkdir(target, ignore_existing=True)
        for item in os.listdir(source):
            itemfull = os.path.join(source, item)
            if os.path.isfile(itemfull):
                self.put(itemfull, '%s/%s' % (target, item))
            else:
                if item in exclude:
                    continue
                self.mkdir('%s/%s' % (target, item), ignore_existing=True)
                self.put_dir(itemfull, '%s/%s' % (target, item))

    def mkdir(self, path, mode=o777, ignore_existing=False):
        """ Augments mkdir by adding an option to not fail if the folder exists  """
        try:
            super(DirSFTPClient, self).mkdir(path, mode)
        except IOError:
            if ignore_existing:
                pass
            else:
                raise


logger = PhialLogger()


class Phial:
    def __init__(self, ssh_service):
        self.__ssh = ssh_service  # type: SSHClient

    def scp(self, localpath, remotepath):
        if os.path.isfile(localpath):
            logger.info('Coping local file: %s to remote location: %s', localpath, remotepath)
            sftp = self.__ssh.open_sftp()  # type: SFTPClient
            sftp.put(localpath, remotepath)
            sftp.close()
        if os.path.isdir(localpath):
            logger.info('Uploading local directory: %s to remote location: %s', localpath, remotepath)
            self.__put_dir(localpath, remotepath)

    def exec(self, command, capture=False):
        # type: (str, bool) -> (int, str, str)
        logger.info("Executing command: %s", command)
        channel = self.__ssh.get_transport().open_session()  # type: Channel
        channel.get_pty()
        if capture:
            handler = Phial.CaptureOutputHandler()
        else:
            handler = Phial.PrintOutputHandler()
        reader = self.OutputReader(channel, handler)
        channel.exec_command(self._bashify(command))
        while True:
            if channel.exit_status_ready():
                reader.read_all()
                break
            rl, wl, xl = select.select([channel], [], [], 0.0)
            if len(rl) > 0:
                reader.read_chunk()
        if capture:
            return channel.exit_status, handler.collected_output(), handler.collected_error_output()
        else:
            return channel.exit_status, '', ''

    def __put_dir(self, source, target):
        transport = self.__ssh.get_transport()
        with DirSFTPClient.from_transport(transport) as sftp:
            sftp.put_dir(source, target)
            sftp.close()

    @six.add_metaclass(abc.ABCMeta)
    class OutputHandler:
        @abc.abstractmethod
        def out(self, data):
            # type: (str) -> None
            pass

        @abc.abstractmethod
        def err(self, data):
            # type: (str) -> None
            pass

    class CaptureOutputHandler(OutputHandler):
        def __init__(self):
            self.outbuf = b''
            self.errbuf = b''

        def out(self, data):
            self.outbuf += data

        def err(self, data):
            self.errbuf += data

        def collected_output(self):
            return self.outbuf.decode(errors='ignore')

        def collected_error_output(self):
            return self.errbuf.decode(errors='ignore')

    class PrintOutputHandler(OutputHandler):
        def __init__(self):
            colorama.init(strip=False)
            self.outbuf = Phial.OutputBuffer()  # type: Phial.OutputBuffer
            self.errbuf = Phial.OutputBuffer()  # type: Phial.OutputBuffer

        def out(self, data):
            self.outbuf.recv(data)
            lines = self.outbuf.lines_collected()
            for line in lines:
                logger.stdout(line)

        def err(self, data):
            self.errbuf.recv(data)
            lines = self.errbuf.lines_collected()
            for line in lines:
                logger.stderr(line)

    class OutputReader:
        def __init__(self, channel, handler):
            self.__channel = channel  # type: Channel
            self.__handler = handler  # type: Phial.OutputHandler

        def read_all(self):
            while self.__channel.recv_stderr_ready() or self.__channel.recv_ready():
                self.read_chunk(1024)

        def read_chunk(self, size=32):
            if self.__channel.recv_stderr_ready():
                data = self.__channel.recv_stderr(size)
                self.__handler.err(data)
            if self.__channel.recv_ready():
                data = self.__channel.recv(size)
                self.__handler.out(data)

    class OutputBuffer:
        def __init__(self):
            self.buf = b''

        def recv(self, data):
            self.buf += data

        def lines_collected(self):
            # type: () -> Sequence[str]
            splited = self.buf\
                .decode(errors='ignore')\
                .splitlines(keepends=True)
            collected = []
            newbuf = ''
            for line in splited:
                if line.endswith(('\n', '\r\n')):
                    collected.append(line.rstrip())
                else:
                    newbuf += line
            self.buf = newbuf.encode()
            return collected

    def _bashify(self, command):
        filtered = command.replace("'", "\\'")
        return "bash -leo pipefail -c '{command}'".format(command=filtered)


@pytest.fixture
def docker_compose_file(pytestconfig, sut):
    return os.path.join(
        str(pytestconfig.rootdir),
        'integration_tests', 'system-under-tests', sut,
        'docker-compose.yml'
    )


@pytest.fixture
def ssh_service(docker_ip, docker_services):
    """Ensure that "SSH service" is up and responsive."""
    ssh = paramiko.SSHClient()
    port = docker_services.port_for('sut', 22)
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(
        hostname=docker_ip,
        username='root',
        password='phial',
        port=port
    )
    return ssh


def execute(command, success_codes=(0,)):
    """Run a shell command."""
    try:
        output = subprocess.check_output(
            command, stderr=subprocess.STDOUT, shell=True, env=os.environ.copy()
        )
        status = 0
    except subprocess.CalledProcessError as error:
        output = error.output or b''
        status = error.returncode
        command = error.cmd
    output = output.decode(errors='ignore')
    if status not in success_codes:
        raise Exception(
            'Command %r returned %d: """%s""".' % (command, status, output)
        )
    return output


def execute_streaming(command, success_codes=(0,)):
    """Run a shell command streaming output to console."""
    handler = Phial.PrintOutputHandler()

    p = subprocess.Popen(command, shell=True,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    while True:
        retcode = p.poll()
        if retcode is not None:
            handler.err(p.stderr.read())
            handler.out(p.stdout.read())
            break
        while len(select.select([p.stderr], [], [], 0)[0]) == 1:
            data = p.stderr.read(1)
            handler.err(data)
        while len(select.select([p.stdout], [], [], 0)[0]) == 1:
            data = p.stdout.read(1)
            handler.out(data)
        time.sleep(0.01)
    if retcode not in success_codes:
        raise Exception(
            'Command %r returned %d.' % (command, retcode)
        )


@pytest.fixture(scope='session')
def docker_ip():
    """Determine IP address for TCP connections to Docker containers."""

    # When talking to the Docker daemon via a UNIX socket, route all TCP
    # traffic to docker containers via the TCP loopback interface.
    docker_host = os.environ.get('DOCKER_HOST', '').strip()
    if not docker_host:
        return '127.0.0.1'

    match = re.match('^tcp://(.+?):\d+$', docker_host)
    if not match:
        raise ValueError(
            'Invalid value for DOCKER_HOST: "%s".' % (docker_host,)
        )
    return match.group(1)


@attr.s(frozen=True)
class Services(object):
    """."""

    _docker_compose = attr.ib()
    _docker_allow_fallback = attr.ib(default=False)

    _services = attr.ib(init=False, default=attr.Factory(dict))

    def port_for(self, service, port):
        """Get the effective bind port for a service."""

        # Return the container port if we run in no Docker mode.
        if self._docker_allow_fallback:
            return port

        # Lookup in the cache.
        cache = self._services.get(service, {}).get(port, None)
        if cache is not None:
            return cache

        output = self._docker_compose.execute(
            'port %s %d' % (service, port,)
        )
        endpoint = output.strip()
        if not endpoint:
            raise ValueError(
                'Could not detect port for "%s:%d".' % (service, port)
            )

        # Usually, the IP address here is 0.0.0.0, so we don't use it.
        match = int(endpoint.split(':', 1)[1])

        # Store it in cache in case we request it multiple times.
        self._services.setdefault(service, {})[port] = match

        return match

    @staticmethod
    def wait_until_responsive(check, timeout, pause,
                              clock=timeit.default_timer):
        """Wait until a service is responsive."""

        ref = clock()
        now = ref
        while (now - ref) < timeout:
            if check():
                return
            time.sleep(pause)
            now = clock()

        raise Exception(
            'Timeout reached while waiting on service!'
        )


def str_to_list(arg):
    if isinstance(arg, (list, tuple)):
        return arg
    return [arg]


@attr.s(frozen=True)
class DockerComposeExecutor(object):
    _compose_files = attr.ib(convert=str_to_list)
    _compose_project_name = attr.ib()

    def execute(self, subcommand):
        return execute(self.__compose_command(subcommand))

    def execute_streaming(self, subcommand):
        return execute_streaming(self.__compose_command(subcommand))

    def __compose_command(self, subcommand):
        command = "docker-compose"
        for compose_file in self._compose_files:
            command += ' -f "{}"'.format(compose_file)
        command += ' -p "{}" {}'.format(self._compose_project_name, subcommand)
        return command


@pytest.fixture
def docker_compose_project_name(request):
    """ Generate a project name using the current process' PID.

    Override this fixture in your tests if you need a particular project name.
    """
    testname = request.node.name
    return "phial-{testname}".format(testname=testname)


@pytest.fixture
def docker_allow_fallback():
    """Return if want to run against localhost when docker is not available.

    Override this fixture to return `True` if you want the ability to
    run without docker.

    """
    return False


@pytest.fixture
def docker_services(
    docker_compose_file, docker_allow_fallback, docker_compose_project_name, capsys
):
    """Ensure all Docker-based services are up and running."""

    docker_compose = DockerComposeExecutor(
        docker_compose_file, docker_compose_project_name
    )

    # If we allowed to run without Docker, check it's presence
    if docker_allow_fallback is True:
        try:
            execute('docker ps')
        except Exception:
            # Run against localhost
            yield Services(docker_compose, docker_allow_fallback=True)
            return

    # Spawn containers.
    with capsys.disabled():
        docker_compose.execute_streaming('up --build -d')

    try:
        # Let test(s) run.
        yield Services(docker_compose)

    finally:
        # Clean up.
        with capsys.disabled():
            docker_compose.execute_streaming('down -v')



from abc import ABCMeta, abstractmethod

from six import with_metaclass
from typing import Sequence

from puppeter.domain.model.configurer import Configurer
from puppeter.domain.model.csr import CsrAttributesConfiguration
from puppeter.domain.model.installer import Mode


class CsrAttributesSetterGateway(with_metaclass(ABCMeta)):
    @abstractmethod
    def save_csr_attributes(self, csr_attrs, mode):
        # type: (CsrAttributesConfiguration, Mode) -> Sequence[Configurer]
        pass



import re
from os.path import dirname

puppeter_source = dirname(dirname(__file__))
remote_dir = '/usr/src/puppeter-source'


class PuppeterAcceptance:

    PUPPET_VER_3 = re.compile('^3\.\d+\.\d+$')
    PUPPET_VER_4 = re.compile('^4\.\d+\.\d+$')
    PUPPET_VER_5 = re.compile('^5\.\d+\.\d+$')

    def __init__(self, phial):
        self.__phial = phial

    def install_puppeter(self):
        self.__phial.scp(puppeter_source, remote_dir)

        exitcode = self.__script('install-develop.sh')

        assert exitcode == 0

    def run_puppeter(self, answers):
        answersfile = '%s/integration_tests/answers/%s' % (remote_dir, answers)
        exitcode = self.__script('execute.sh', answersfile)

        assert exitcode == 0

    def __script(self, script, arg=''):
        command = 'bash -leo pipefail %s/integration_tests/scripts/%s %s' % (remote_dir, script, arg)
        exitcode, out, err = self.__phial.exec(command.strip())
        return exitcode



def merge(*dict_args):
    """
    Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.
    """
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result



from puppeter.container import Named
from puppeter.domain.facter import Facter
from puppeter.domain.model.ordered import Order
from puppeter.domain.model.osfacts import OperatingSystemCodename
from puppeter.persistence.gateway.installer.debian.after4x import AfterPuppet4xConfigurer


@Named('pc4x-debian')
@Order(100)
class DebianPC4xConfigurer(AfterPuppet4xConfigurer):

    def __init__(self, installer):
        AfterPuppet4xConfigurer.__init__(self, installer)

    def _collect_repo(self, collector):
        codename = Facter.get(OperatingSystemCodename)
        collector.collect_from_template(
            'Puppet Labs Collection Repository (PC1) setup (Puppet OSS 4.x)',
            'pc4x-repo.sh',
            dict(codename=codename)
        )



from puppeter.domain.facter import Facter


class MockFacter(Facter):
    @staticmethod
    def reset_fact(enumvar):
        MockFacter._facts.pop(enumvar, None)

    @staticmethod
    def reset_facts():
        MockFacter._facts.clear()

    @staticmethod
    def set_fact(enumvar, value):
        MockFacter._facts[enumvar] = value



from typing import Sequence, MutableSequence

from puppeter.container import Named
from puppeter.domain.facter import Facter
from puppeter.domain.model import Installer
from puppeter.domain.model.configurer import Configurer, ScriptFormat
from puppeter.domain.model.installer import Mode, After4xCollectionInstaller
from puppeter.domain.model.javafacts import JavaVersion
from puppeter.domain.model.ordered import Order
from puppeter.domain.model.osfacts import Docker


@Named('puppet.service.puppetserver-service')
@Order(1000)
class PuppetServerServiceStarterConfigurer(Configurer):
    def __init__(self, installer):
        self.__installer = installer  # type: Installer

    def produce_commands(self):
        if self.__installer.mode() == Mode.Agent:
            return []
        if self.__installer.is_after_4x():
            servicename = 'puppetserver'
        else:
            servicename = 'puppetmaster'
        enable = 'true' if Facter.get(Docker) is not Docker.YES else 'false'
        collector = self._collector()
        desc = 'Starting Puppet server service ({servicename})'.format(servicename=servicename)
        mapping = dict(servicename=servicename, enable=enable)
        collector.collect_from_template(desc, 'puppetserver.sh', mapping)
        return collector.lines()


@Named('puppet.service.puppetserver-jvmargs')
@Order(900)
class PuppetServerJvmArgsConfigurer(Configurer):
    def __init__(self, installer):
        self.__installer = installer  # type: Installer

    def produce_commands(self):
        install = self.__installer
        if not (isinstance(install, After4xCollectionInstaller) and
                install.mode() == Mode.Server and
                install.is_after_4x()):
            # Only for Java written PuppetServer
            return []
        return self.__produce_commands(install)

    def __produce_commands(self, install):
        # type: (After4xCollectionInstaller) -> Sequence[str]
        args = []  # type: MutableSequence[str]
        mem = install.puppetserver_jvm_memory()
        jvmargs = install.puppetserver_jvm_args()
        if mem.is_set():
            if mem.heap_minimum() is not None:
                args.append('-Xms{xms}'.format(xms=mem.heap_minimum()))
            if mem.heap_maximum() is not None:
                args.append('-Xmx{xmx}'.format(xmx=mem.heap_maximum()))
            java_version = Facter.get(JavaVersion)
            metaspace_arg = self.__get_metaspace_arg(java_version)
            if java_version.has_permgen_space() and mem.metaspace_maximum() is None:
                args.append(metaspace_arg.format(mspace='256m'))
            elif mem.metaspace_maximum() is not None:
                args.append(metaspace_arg.format(mspace=mem.metaspace_maximum()))
        if jvmargs.are_set():
            args.extend(jvmargs)
        args_as_str = ' '.join(args)
        collector = self._collector()
        mapping = dict(jvmargs=args_as_str)
        collector.collect_from_template('Configuring PuppetServer JVM Args',
                                        'puppetserver-jvmargs.pp',
                                        mapping,
                                        format=ScriptFormat.PUPPET)
        return collector.lines()

    def __get_metaspace_arg(self, java_version):
        # type: (JavaVersion) -> str
        return '-XX:MaxPermSize={mspace}' if java_version.has_permgen_space() else '-XX:MaxMetaspaceSize={mspace}'



from puppeter.domain.model.csr import CsrAttributesConfiguration
from puppeter.domain.model.installer import Mode
from puppeter.persistence.gateway.csr import CsrAttributesConfigurer, CsrAttributesSetterGatewayImpl
from puppeter import container


def test_csr_on_server():
    # given
    config = CsrAttributesConfiguration()
    config.read_raw_options(dict(
        pp_role='appserver'
    ))
    configurer = CsrAttributesConfigurer(config, Mode.Server)
    container.initialize()
    # when
    commands = configurer.produce_commands()

    # then
    assert '  pp_role: appserver' in commands
    assert '  group   => \'puppet\',' in commands


def test_csr_on_agent():
    # given
    config = CsrAttributesConfiguration()
    config.read_raw_options(dict(
        pp_role='agent'
    ))
    configurer = CsrAttributesConfigurer(config, Mode.Agent)
    container.initialize()

    # when
    commands = configurer.produce_commands()

    # then
    assert '  pp_role: agent' in commands


def test_csr_gateway():
    # given
    gateway = CsrAttributesSetterGatewayImpl()
    config = CsrAttributesConfiguration()
    config.read_raw_options(dict(
        pp_datacenter='dc1'
    ))
    mode = Mode.Server

    # when
    configurers = gateway.save_csr_attributes(config, mode)

    # then
    assert len(configurers) == 1
    op = getattr(configurers[0], "produce_commands", None)
    assert op is not None
    assert callable(op)



from os.path import dirname, join
import ruamel.yaml

from puppeter.domain.model.answers import Answers
from puppeter.domain.model.fqdn import FqdnConfiguration
from puppeter.domain.model.installer import Collection4xInstaller, Mode
from puppeter.persistence.gateway.answers import YamlAnswersGateway

EXAMPLE_ANSWERS = open(join(dirname(__file__), 'example-answers-simple.yaml'), mode='r').read()


def __load(yaml_content):
    return ruamel.yaml.load(yaml_content, Loader=ruamel.yaml.Loader)


def test_write(tmpdir):
    # given
    installer = Collection4xInstaller()
    installer.read_raw_options({
        'mode': 'Server'
    })
    answers = Answers(
        installer=installer,
        fqdn_configuration=FqdnConfiguration('app6.acme.internal')
    )
    gateway = YamlAnswersGateway()
    target = tmpdir.join('answers.yaml')

    # when
    gateway.write_answers_to_file(answers, target)

    # then
    assert __load(target.read()) == __load(EXAMPLE_ANSWERS)


def test_read(tmpdir):
    # given
    gateway = YamlAnswersGateway()
    target = tmpdir.join('answers.yaml')
    target.write(EXAMPLE_ANSWERS)

    # when
    answers = gateway.read_answers_from_file(target)

    # then
    assert answers is not None
    assert isinstance(answers, Answers)
    assert isinstance(answers.installer(), Collection4xInstaller)
    assert answers.installer().mode() == Mode.Server
    assert answers.fqdn_configuration().hostname() == 'app6'
    assert answers.fqdn_configuration().has_domain()



from typing import TypeVar, Type, Callable, Optional

T = TypeVar('T')


class Facter:
    """
    A nano version of Puppet's Facter
    """
    _facts = {}
    __resolvers = {}

    @staticmethod
    def get(enumvar):
        # type: (Type[T]) -> Optional[T]
        if enumvar in Facter._facts:
            return Facter._facts[enumvar]
        elif enumvar in Facter.__resolvers:
            resolved = Facter.__resolve(enumvar)
            if resolved is not None:
                Facter._facts[enumvar] = resolved
            return resolved
        else:
            return None

    @staticmethod
    def set(enumvar, resolver):
        # type: (Type[T], Callable[[], T]) -> None
        if enumvar not in Facter.__resolvers:
            Facter.__resolvers[enumvar] = []
        Facter.__resolvers[enumvar].append(resolver)

    @staticmethod
    def __resolve(enumvar):
        # type: (Type[T]) -> Optional[T]
        for resolver in Facter.__resolvers[enumvar]:
            resolved = resolver()
            if resolved is not None:
                return resolved
        return None



import logging

from logging import Logger
from typing import Type

# The version of the app
__version__ = '0.8.0'
__program__ = 'puppeter'


def get_logger(cls):
    # type: (Type) -> Logger
    name = __fullname(cls)
    return logging.getLogger(name)


def __fullname(cls):
    return cls.__module__ + "." + cls.__name__


if __name__ == '__main__':
    from puppeter.main import main
    main()



import sys
from contextlib import contextmanager
from io import BytesIO, StringIO


@contextmanager
def captured_output():
    if sys.version_info >= (3, 0):
        new_out, new_err = StringIO(), StringIO()
    else:
        new_out, new_err = BytesIO(), BytesIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err



from puppeter import container
from puppeter.domain.model import SystemInstaller
from puppeter.domain.model.osfacts import OsFamily, OperatingSystem, OperatingSystemRelease
from puppeter.persistence.gateway.installer.redhat import RedHatSystemConfigurer
from tests.domain.mock_facter import MockFacter


def test_system_redhat():
    # given
    MockFacter.set_fact(OperatingSystem, OperatingSystem.RedHat)
    MockFacter.set_fact(OperatingSystemRelease, OperatingSystemRelease('7.3.11'))
    installer = SystemInstaller()
    installer.read_raw_options({'mode': 'Server'})
    configurer = RedHatSystemConfigurer(installer=installer)

    # when
    commands = configurer.produce_commands()
    commands = list(map(lambda cmd: cmd.strip(), commands))

    # then
    assert 'yum install -y wget' not in commands
    assert "rpm -Uvh 'https://dl.fedoraproject.org/pub/epel/epel-release-latest-7.noarch.rpm'" in commands
    assert 'yum install -y puppet' in commands
    assert 'puppet resource package puppetmaster ensure=installed' in commands


def setup_function():
    MockFacter.set_fact(OsFamily, OsFamily.RedHat)
    container.initialize()


def teardown_function():
    MockFacter.reset_facts()



from puppeter import container
from puppeter.container import Named
from puppeter.domain.model.configurer import Configurer
from puppeter.domain.gateway.installer import InstallerGateway
from puppeter.persistence.gateway.installer.linux import LinuxInstallerGateway
from puppeter.persistence.gateway.installer.redhat.pc3x import RedHatPC3xConfigurer
from puppeter.persistence.gateway.installer.redhat.pc4x import RedHatPC4xConfigurer
from puppeter.persistence.gateway.installer.redhat.pc5x import RedHatPC5xConfigurer
from puppeter.persistence.gateway.installer.redhat.system import RedHatSystemConfigurer


@Named('redhat')
class RedHatInstallerGateway(LinuxInstallerGateway):

    def _provide_install_configurers(self, installer):
        configurers = [self.__provide_install_configurer(installer)]
        return tuple(configurers)

    @staticmethod
    def __provide_install_configurer(installer):
        name = installer.bean_name()
        if name == 'gem':
            return container.get_named(Configurer, 'gem', installer=installer)
        name += '-redhat'
        return container.get_named(Configurer, name, installer=installer)


container.bind(InstallerGateway, RedHatInstallerGateway)
container.bind(Configurer, RedHatSystemConfigurer)
container.bind(Configurer, RedHatPC3xConfigurer)
container.bind(Configurer, RedHatPC4xConfigurer)
container.bind(Configurer, RedHatPC5xConfigurer)



from six import with_metaclass
from abc import ABCMeta, abstractmethod

from typing import IO

from puppeter.domain.model.answers import Answers


class AnswersGateway(with_metaclass(ABCMeta, object)):
    @abstractmethod
    def read_answers_from_file(self, target):
        # type: (IO) -> Answers
        pass

    @abstractmethod
    def write_answers_to_file(self, answers, target):
        # type: (Answers, IO) -> None
        pass


class AnswersProcessor(with_metaclass(ABCMeta, object)):

    @abstractmethod
    def process(self, answers):
        # type: (Answers) -> None
        pass



from __future__ import absolute_import

from six import with_metaclass
from abc import ABCMeta, abstractmethod


class InstallerGateway(with_metaclass(ABCMeta, object)):

    def configurers(self, installer):
        configurers = []
        if installer is not None:
            configurers.extend(self._provide_install_configurers(installer))
            configurers.extend(self._puppet_cert_issue(installer))
            configurers.extend(self._puppet_conf(installer))
            configurers.extend(self._puppet_services(installer))
        return tuple(configurers)

    @abstractmethod
    def _provide_install_configurers(self, installer):
        pass

    @abstractmethod
    def _puppet_cert_issue(self, installer):
        pass

    @abstractmethod
    def _puppet_services(self, installer):
        pass

    @abstractmethod
    def _puppet_conf(self, installer):
        pass



from puppeter.container import Named
from puppeter.domain.model.ordered import Order
from puppeter.persistence.gateway.installer.redhat.after4x import AfterPuppet4xConfigurer


@Named('pc4x-redhat')
@Order(100)
class RedHatPC4xConfigurer(AfterPuppet4xConfigurer):

    def __init__(self, installer):
        AfterPuppet4xConfigurer.__init__(self, installer)

    def _repo_script_path_el5(self):
        return 'pc4x-el5-repo.sh'

    def _repo_script_path(self):
        return 'pc4x-repo.sh'

    def _repo_setup_title(self):
        return 'Puppet Labs Collection Repository (PC1) setup (Puppet OSS 4.x)'



import logging
import os
import sys
from abc import abstractmethod, ABCMeta
from logging import StreamHandler
from logging.handlers import SysLogHandler
from typing import IO, Any

from six import with_metaclass

import puppeter
from puppeter import container
from puppeter.domain.facter import Facter
from puppeter.domain.gateway.answers import AnswersProcessor
from puppeter.domain.model.answers import Answers
from puppeter.domain.model.osfacts import OsFamily


class Options:
    def __init__(self, namespace):
        self.__answers = namespace.answers  # type: IO[Any]
        self.__verbose = namespace.verbose  # type: int
        self.__execute = namespace.execute  # type: bool

    def answers(self):
        # type: () -> IO[Any]
        return self.__answers

    def verbose(self):
        # type: () -> int
        return self.__verbose

    def execute(self):
        # type: () -> bool
        return self.__execute


class App(with_metaclass(ABCMeta, object)):

    def __init__(self, options):
        self._options = options  # type: Options

    def run(self):
        root = logging.getLogger()
        level = self.__get_log_level(self._options.verbose())
        root.setLevel(level)
        handlers = (self.__syslog_handler(), self.__stderr_handler())
        for hndl in handlers:
            root.addHandler(hndl)
        answers = self._collect_answers()
        self.__process(answers)

    @abstractmethod
    def _collect_answers(self):
        # type: () -> Answers
        pass

    def __process(self, answers):
        # type: (Answers) -> None
        processor = container.get(AnswersProcessor, options=self._options)  # type: AnswersProcessor
        processor.process(answers)

    @staticmethod
    def __stderr_handler():
        # type: () -> logging.Handler
        handler = StreamHandler(stream=sys.stderr)
        fmt = '# %(levelname)s: %(message)s'
        handler.setFormatter(ColoredFormatter(fmt=fmt))
        handler.setLevel(logging.NOTSET)
        return handler

    @staticmethod
    def __syslog_handler():
        # type: () -> logging.Handler
        osfamily = Facter.get(OsFamily)
        if osfamily in (OsFamily.Debian, OsFamily.RedHat, OsFamily.Suse):
            try:
                handler = SysLogHandler(address='/dev/log')
            except EnvironmentError:
                handler = SysLogHandler()
        else:
            handler = SysLogHandler()
        fmt = puppeter.__program__ + '[%(process)d]: %(levelname)s %(name)s - %(message)s'
        handler.setFormatter(logging.Formatter(fmt=fmt))
        handler.setLevel(logging.INFO)
        return handler

    @staticmethod
    def __get_log_level(verbosity):
        if verbosity is None:
            verbosity = 0
        if verbosity > 2:
            verbosity = 2
        levels = {
            0: logging.WARNING,
            1: logging.INFO,
            2: logging.DEBUG
        }
        return levels[verbosity]


def color_code(fg, bg=None):
    if bg is not None:
        return FGBG_COLOR_SEQ % (fg, bg)
    else:
        return FG_COLOR_SEQ % fg


RESET_SEQ = "\033[0m"
FG_COLOR_SEQ = "\033[38;5;%dm"
FGBG_COLOR_SEQ = "\033[38;5;%d;48;5;%dm"
# ref: https://unix.stackexchange.com/a/124409/145501
COLORS = {
    'WARNING': color_code(fg=226),
    'INFO': color_code(fg=155),
    'DEBUG': color_code(fg=244),
    'CRITICAL': color_code(fg=219, bg=196),
    'ERROR': color_code(fg=196)
}
try:
    SUPPORTS_256_COLORS = os.environ['XTERM_256_COLORS'] == '1'
except KeyError:
    SUPPORTS_256_COLORS = False


class ColoredFormatter(logging.Formatter):

    def __init__(self, fmt, use_color=SUPPORTS_256_COLORS):
        logging.Formatter.__init__(self, fmt=fmt)
        self.use_color = use_color

    def format(self, record):
        formatted = logging.Formatter.format(self, record)
        levelname = record.levelname
        if self.use_color and levelname in COLORS:
            formatted = COLORS[levelname] + formatted + RESET_SEQ
        return formatted



import re

import puppeter.settings
from puppeter import container
from puppeter.domain.facter import Facter
from puppeter.domain.model.osfacts import OsFamily, OperatingSystem, OperatingSystemRelease, OperatingSystemCodename, \
    Docker
from tests.domain.mock_facter import MockFacter


class TestOsFactsResolvingOnMocks:

    # noinspection PyMethodMayBeStatic
    def setup_method(self):
        MockFacter.set_fact(OperatingSystemCodename, OperatingSystemCodename('trusty'))
        MockFacter.set_fact(OperatingSystemRelease, OperatingSystemRelease('14.04'))
        MockFacter.set_fact(OperatingSystem, OperatingSystem.Ubuntu)
        MockFacter.set_fact(OsFamily, OsFamily.Debian)
        container.initialize()

    # noinspection PyMethodMayBeStatic
    def teardown_method(self):
        MockFacter.reset_facts()

    def test_osfamily(self):
        osfamily = Facter.get(OsFamily)

        assert osfamily is not OsFamily.Unknown
        assert osfamily is not None

    def test_operatingsystem(self):
        os = Facter.get(OperatingSystem)

        assert os is not OperatingSystem.Unknown
        assert os is not None

    def test_oscodename(self):
        codename = Facter.get(OperatingSystemCodename)

        assert codename == 'trusty'
        assert type(codename) is OperatingSystemCodename

    def test_osrelease(self):
        regex = OperatingSystemRelease.VERSION_RE
        rel = Facter.get(OperatingSystemRelease)

        assert rel is not None
        assert regex.match(rel) is not None

    def test_osrelrelease_major(self):
        rel = Facter.get(OperatingSystemRelease)
        major = rel.major()

        assert major is not None
        assert type(major) is str
        assert re.compile('^\d+$').match(major) is not None

    def test_osrelrelease_minor(self):
        rel = Facter.get(OperatingSystemRelease)
        minor = rel.minor()

        assert minor is not None
        assert type(minor) is str
        assert re.compile('^\d+$').match(minor) is not None

    def test_osrelease_major_oninvalid(self):
        MockFacter.set_fact(OperatingSystemRelease, OperatingSystemRelease('alpha'))

        rel = Facter.get(OperatingSystemRelease)
        major = rel.major()

        assert major is not None
        assert major is 'alpha'


class TestOsFactsResolvingForReal:
    @classmethod
    def setup_class(cls):
        container.initialize()

    # noinspection PyMethodMayBeStatic
    def teardown_method(self):
        MockFacter.reset_facts()

    def test_osfamily(self):
        osfamily = Facter.get(OsFamily)

        assert osfamily is not None

    def test_operatingsystem(self):
        os = Facter.get(OperatingSystem)

        assert os is not None

    def test_oscodename(self):
        codename = Facter.get(OperatingSystemCodename)

        assert codename is not None

    def test_osrelease(self):
        rel = Facter.get(OperatingSystemRelease)

        assert rel is not None

    def test_docker(self, tmpdir):
        old_chroot = puppeter.settings.chroot
        try:
            new_root = tmpdir.mkdir("docker-utest")
            puppeter.settings.chroot = str(new_root) + '/'
            p = new_root.mkdir('proc').mkdir('1').join('cgroup')
            p.write('11:blkio:/docker/885762619\n'
                    '10:pids:/docker/885762619'
                    '9:freezer:/docker/885762619\n'
                    '8:cpuset:/docker/885762619\n'
                    '7:perf_event:/docker/885762619\n'
                    '6:devices:/docker/885762619\n'
                    '5:hugetlb:/docker/885762619\n')
            docker = Facter.get(Docker)

            assert type(docker) is Docker
            assert docker is Docker.YES
            assert bool(docker) is True
        finally:
            puppeter.settings.chroot = old_chroot



from enum import Enum
from six import iteritems

from puppeter.domain.model.withoptions import WithOptions


class PuppetOID(Enum):
    pp_uuid = 1
    pp_instance_id = 2
    pp_image_name = 3
    pp_preshared_key = 4
    pp_cost_center = 5
    pp_product = 6
    pp_project = 7
    pp_application = 8
    pp_service = 9
    pp_employee = 10
    pp_created_by = 11
    pp_environment = 12
    pp_role = 13
    pp_software_version = 14
    pp_department = 15
    pp_cluster = 16
    pp_provisioner = 17
    pp_region = 18
    pp_datacenter = 19
    pp_zone = 20
    pp_network = 21
    pp_securitypolicy = 22
    pp_cloudplatform = 23
    pp_apptier = 24
    pp_hostname = 25


class CsrAttributesConfiguration(WithOptions):

    def __init__(self):
        self.__csr = {}

    def raw_options(self):
        raw = {}
        for (poid, value) in iteritems(self.__csr):
            raw[poid.name] = value
        return raw

    def read_raw_options(self, options):
        self.__csr.clear()

        for (key, val) in iteritems(options):
            try:
                poid = PuppetOID[key]
                self.__csr[poid] = val
            except KeyError:
                raise KeyError('Unknown Puppet OID for CSR: {}', key)



from abc import ABCMeta, abstractmethod

from six import with_metaclass

from puppeter.domain.facter import Facter
from puppeter.domain.model.osfacts import OperatingSystemRelease, OperatingSystem
from puppeter.persistence.gateway.installer.redhat.base import BaseRedHatConfigurer


class AfterPuppet4xConfigurer(with_metaclass(ABCMeta, BaseRedHatConfigurer)):
    def __init__(self, installer):
        super(AfterPuppet4xConfigurer, self).__init__(installer)

    def _collect_agent(self, collector):
        collector.collect_from_file('Puppet Agent setup', 'puppetagent.sh')

    def _collect_server(self, collector):
        collector.collect_from_file('Puppet Server setup', 'puppetserver.sh')

    def _collect_repo(self, collector):
        major = Facter.get(OperatingSystemRelease).major()
        os = Facter.get(OperatingSystem)
        if os == OperatingSystem.Fedora:
            abbr = 'fedora'
        else:
            abbr = 'el'
        mapping = dict(major=major, abbr=abbr)
        if ('%s-%s' % (abbr, major)) == 'el-5':
            script = self._repo_script_path_el5()
        else:
            script = self._repo_script_path()
        collector.collect_from_template(self._repo_setup_title(), script, mapping)

    @abstractmethod
    def _repo_setup_title(self):
        raise NotImplementedError()

    @abstractmethod
    def _repo_script_path_el5(self):
        raise NotImplementedError()

    @abstractmethod
    def _repo_script_path(self):
        raise NotImplementedError()



import pytest
from os.path import join

from integration_tests.acceptance import PuppeterAcceptance


@pytest.fixture
def sut():
    """System Under Test"""
    return join('debian', 'ubuntu-1604')


@pytest.mark.ubuntu1604
def test_agent_system_on_ubuntu_1604(phial, capsys, regex):
    with capsys.disabled():
        acceptance = PuppeterAcceptance(phial)

        acceptance.install_puppeter()
        acceptance.run_puppeter('agent-system.yml')

        exitcode, output, error = phial.exec('puppet --version', capture=True)
        assert error == ''
        regex.pattern(PuppeterAcceptance.PUPPET_VER_3).matches(output)
        assert exitcode == 0


@pytest.mark.ubuntu1604
def test_agent_pc4x_on_ubuntu_1604(phial, capsys, regex):
    with capsys.disabled():
        acceptance = PuppeterAcceptance(phial)

        acceptance.install_puppeter()
        acceptance.run_puppeter('agent-pc4x.yml')

        exitcode, output, error = phial.exec('puppet --version', capture=True)
        assert error == ''
        regex.pattern(PuppeterAcceptance.PUPPET_VER_4).matches(output)
        assert exitcode == 0


@pytest.mark.ubuntu1604
def test_agent_pc5x_on_ubuntu_1604(phial, capsys, regex):
    with capsys.disabled():
        acceptance = PuppeterAcceptance(phial)

        acceptance.install_puppeter()
        acceptance.run_puppeter('agent-pc5x.yml')

        exitcode, output, error = phial.exec('puppet --version', capture=True)
        assert error == ''
        regex.pattern(PuppeterAcceptance.PUPPET_VER_5).matches(output)
        assert exitcode == 0



from puppeter import container
from puppeter.domain.model.configurer import CommandsCollector, Configurer
from puppeter.domain.model.javafacts import JavaVersion
from puppeter.persistence.service.commandscollector import CommandsCollectorImpl
from puppeter.persistence.service.java import calculate_java_version
from puppeter.persistence.service.os import calculate_operatingsystem, \
    calculate_osfamily, calculate_osrelease, calculate_oscodename, calculate_docker
from puppeter.domain.facter import Facter
from puppeter.domain.model.osfacts import OperatingSystem, OsFamily, \
    OperatingSystemRelease, OperatingSystemCodename, Docker
from puppeter.persistence.service.puppetconf import PuppetConfConfigurer
from puppeter.persistence.service.puppetserver import PuppetServerServiceStarterConfigurer, \
    PuppetServerJvmArgsConfigurer

Facter.set(OperatingSystem, calculate_operatingsystem)
Facter.set(OsFamily, calculate_osfamily)
Facter.set(OperatingSystemRelease, calculate_osrelease)
Facter.set(OperatingSystemCodename, calculate_oscodename)
Facter.set(Docker, calculate_docker)

Facter.set(JavaVersion, calculate_java_version)

container.bind(CommandsCollector, CommandsCollectorImpl)
container.bind(Configurer, PuppetServerServiceStarterConfigurer)
container.bind(Configurer, PuppetServerJvmArgsConfigurer)
container.bind(Configurer, PuppetConfConfigurer)



import re
from abc import abstractmethod, ABCMeta

from enum import Enum
from six import with_metaclass
from typing import Sequence

from puppeter.container import Named
from puppeter.domain.model.gemrequirement import GemRequirement
from puppeter.domain.model.withoptions import WithOptions
from puppeter.util.dict import merge


class Mode(Enum):
    Agent = 1
    Server = 2


class Installer(WithOptions):
    def __init__(self):
        self.__mode = Mode.Agent  # type: Mode
        self.__puppetconf = PuppetConf()  # type: PuppetConf

    def raw_options(self):
        # noinspection PyUnresolvedReferences
        installer_type = self.bean_name()
        return merge({
            'mode': self.__mode.name,
            'type': installer_type
        }, self.__puppetconf.raw_options())

    def read_raw_options(self, options):
        try:
            self.__mode = Mode[options['mode']]
            self.__puppetconf.read_raw_options(options)
        except KeyError:
            pass

    def mode(self):
        # type: () -> Mode
        return self.__mode

    @abstractmethod
    def is_after_4x(self):
        # type: () -> bool
        pass

    def puppetconf(self):
        # type: () -> PuppetConf
        return self.__puppetconf


@Named('system')
class SystemInstaller(Installer):
    def is_after_4x(self):
        return False

    def __init__(self):
        Installer.__init__(self)


@Named('gem')
class RubygemsInstaller(Installer):
    def is_after_4x(self):
        return GemRequirement(self.version()).satified_by('4.0')

    def __init__(self):
        Installer.__init__(self)
        self.__version = None

    def raw_options(self):
        parent_options = Installer.raw_options(self)
        merged = parent_options.copy()
        merged.update({
            'version': self.__version
        })
        return merged

    def read_raw_options(self, options):
        Installer.read_raw_options(self, options)
        try:
            self.__version = str(options['version'])
        except KeyError:
            pass

    def version(self):
        return self.__version


class CollectionInstaller(with_metaclass(ABCMeta, Installer)):
    pass


class MemScale(Enum):
    KILO = 1
    MEGA = 2
    GIGA = 3

    @staticmethod
    def of(repr):
        return {
            'k': MemScale.KILO,
            'm': MemScale.MEGA,
            'g': MemScale.GIGA
        }[repr]

    def to_s(self):
        return {
            MemScale.KILO: 'k',
            MemScale.MEGA: 'm',
            MemScale.GIGA: 'g'
        }[self]


class JavaMemorySpec:
    def __init__(self, num, scale=MemScale.MEGA):
        self.__num = int(num)  # type: int
        self.__scale = scale  # type: MemScale

    @staticmethod
    def of(repr):
        # https://regex101.com/r/YATBuy/1
        pattern = re.compile('^([0-9]+)([kmg])$', re.IGNORECASE)
        match = pattern.match(repr)
        num = int(match.group(1))
        scale_repr = match.group(2).lower()
        scale = MemScale.of(scale_repr)
        return JavaMemorySpec(num, scale)

    def scale(self):
        return self.__scale

    def number(self):
        return self.__num

    def __str__(self):
        return '%d%s' % (self.__num, self.__scale.to_s())

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return str(self) == str(other)


class JavaMemory(WithOptions):
    def __init__(self, heap_maximum=None, heap_minimum=None, metaspace_maximum=None):
        self.__heap_maximum = heap_maximum  # type: JavaMemorySpec
        if heap_minimum is not None:
            self.__heap_minimum = heap_minimum  # type: JavaMemorySpec
        else:
            self.__heap_minimum = heap_maximum  # type: JavaMemorySpec
        self.__metaspace_maximum = metaspace_maximum  # type: JavaMemorySpec

    def raw_options(self):
        d = {}
        if self.heap_maximum() is not None:
            self._update(d, {'heap': {'max': str(self.heap_maximum())}})
        if self.heap_minimum() is not None:
            self._update(d, {'heap': {'min': str(self.heap_minimum())}})
        if self.metaspace_maximum() is not None:
            self._update(d, {'metaspace': {'max': str(self.metaspace_maximum())}})
        return {
            'puppetserver': {
                'jvm': {'memory': d}
            }
        } if (len(d) > 0) else {}

    def read_raw_options(self, options):
        try:
            mem = options['puppetserver']['jvm']['memory']
        except KeyError:
            return
        try:
            self.__heap_maximum = JavaMemorySpec.of(mem['heap']['max'])
        except KeyError:
            pass
        try:
            self.__heap_minimum = JavaMemorySpec.of(mem['heap']['min'])
        except KeyError:
            if self.__heap_maximum is not None:
                self.__heap_minimum = self.__heap_maximum
        try:
            self.__metaspace_maximum = JavaMemorySpec.of(mem['metaspace']['max'])
        except KeyError:
            pass

    def heap_maximum(self):
        return self.__heap_maximum

    def heap_minimum(self):
        return self.__heap_minimum

    def metaspace_maximum(self):
        return self.__metaspace_maximum

    def is_set(self):
        return self.heap_maximum() is not None \
            or self.heap_minimum() is not None \
            or self.metaspace_maximum() is not None


class JvmArgs(WithOptions, Sequence):

    def __init__(self, args=tuple()):
        self.__args = []
        self.__args.extend(args)

    def __getitem__(self, i):
        return self.__args[i]

    def __len__(self):
        return len(self.__args)

    def raw_options(self):
        return {
            'puppetserver': {
                'jvm': {'args': tuple(self.__args)}
            }
        }

    def read_raw_options(self, options):
        try:
            new_args = options['puppetserver']['jvm']['args']
            del self.__args[:]
            self.__args.extend(new_args)
        except KeyError:
            pass

    def are_set(self):
        return len(self.__args) > 0


class PuppetConf(WithOptions, dict):
    def __init__(self, options=None):
        if options is None:
            options = {}
        super(PuppetConf, self).__init__(options)

    def read_raw_options(self, options):
        try:
            opts = options['puppet.conf']
            self.clear()
            self.update(opts)
        except KeyError:
            pass

    def raw_options(self):
        if len(self) > 0:
            return {'puppet.conf': self}
        else:
            return {}


class After4xCollectionInstaller(CollectionInstaller):
    def __init__(self):
        CollectionInstaller.__init__(self)
        self.__mem = JavaMemory()  # type: JavaMemory
        self.__jvmargs = JvmArgs()  # type: JvmArgs

    def is_after_4x(self):
        return True

    def raw_options(self):
        options = super(CollectionInstaller, self).raw_options()
        if self.__mem.is_set():
            self._update(options, self.__mem.raw_options())
        if self.__jvmargs.are_set():
            self._update(options, self.__jvmargs.raw_options())
        return options

    def read_raw_options(self, options):
        super(CollectionInstaller, self).read_raw_options(options)
        if self.mode() == Mode.Server:
            self.__jvmargs.read_raw_options(options)
            try:
                self.__mem.read_raw_options(options)
            except KeyError:
                pass

    def puppetserver_jvm_memory(self):
        return self.__mem

    def puppetserver_jvm_args(self):
        return self.__jvmargs


@Named('pc3x')
class Collection3xInstaller(CollectionInstaller):
    def is_after_4x(self):
        return False


@Named('pc4x')
class Collection4xInstaller(After4xCollectionInstaller):
    pass


@Named('pc5x')
class Collection5xInstaller(After4xCollectionInstaller):
    pass



from typing import Optional

from puppeter.domain.model.withoptions import WithOptions


class FqdnConfiguration(WithOptions):
    def __init__(self, fqdn='localhost.localdomain'):
        self.__fqdn = fqdn.strip()

    def raw_options(self):
        return self.fqdn()

    def read_raw_options(self, options):
        self.__fqdn = str(options).strip()

    def fqdn(self):
        # type: () -> str
        return self.__fqdn

    def hostname(self):
        # type: () -> str
        return self.fqdn().split('.')[0]

    def domain(self):
        # type: () -> Optional[str]
        domain = '.'.join(self.fqdn().split('.')[1:])
        domain = None if domain == '' else domain
        return domain

    def has_domain(self):
        # type: () -> bool
        return self.domain() is not None



import ruamel.yaml

from typing import Sequence

from puppeter.domain.gateway.csr import CsrAttributesSetterGateway
from puppeter.domain.model.configurer import Configurer, ScriptFormat
from puppeter.domain.model.csr import CsrAttributesConfiguration
from puppeter.domain.model.installer import Mode
from puppeter.domain.model.ordered import Order


@Order(250)
class CsrAttributesConfigurer(Configurer):
    def __init__(self, csr_attrs, mode):
        self.__csr_attr = csr_attrs  # type: CsrAttributesConfiguration
        self.__mode = mode  # type: Mode

    def produce_commands(self):
        mapping = dict(
            content=self.__as_yaml()
        )
        if self.__mode == Mode.Agent:
            template = 'csr_agent.pp'
        elif self.__mode == Mode.Server:
            template = 'csr_server.pp'
        else:
            raise Exception('Invalid mode given: {mode}'.format(mode=self.__mode))

        return self._collector().collect_from_template(
            description='Configuring puppet node certificate CSR attributes',
            template=template,
            mapping=mapping,
            format=ScriptFormat.PUPPET
        ).lines()

    def __as_yaml(self):
        csr = {
            'extension_requests': self.__csr_attr.raw_options()
        }
        return ruamel.yaml.dump(csr, Dumper=ruamel.yaml.RoundTripDumper)


class CsrAttributesSetterGatewayImpl(CsrAttributesSetterGateway):
    def save_csr_attributes(self, csr_attrs, mode):
        # type: (CsrAttributesConfiguration, Mode) -> Sequence[Configurer]
        return [
            CsrAttributesConfigurer(csr_attrs, mode)
        ]



from puppeter.domain.model import Collection4xInstaller
from puppeter.persistence.service import PuppetConfConfigurer


def test_puppetconf():
    # given
    installer = Collection4xInstaller()
    installer.read_raw_options({
        'mode': 'Agent',
        'puppet.conf': {
            'main': {
                'server': 'puppet.acme.internal'
            },
            'agent': {
                'noop': True
            }
        }
    })
    configurer = PuppetConfConfigurer(installer)

    # when
    commands = configurer.produce_commands()

    # then
    assert configurer.bean_name() == 'puppet.conf'
    assert configurer.order() == 900
    assert found("set main/server 'puppet.acme.internal'").inside(commands)
    assert found("set agent/noop true").inside(commands)


def found(st):
    class Found:
        def __init__(self):
            pass

        @staticmethod
        def inside(lst):
            return any(st in s for s in lst)
    return Found



import pytest

from puppeter.domain.model.osfacts import Docker, OsFamily
from puppeter.persistence.gateway.fqdn import FqdnSetterGatewayImpl

from puppeter.domain.model.fqdn import FqdnConfiguration
from tests.domain.mock_facter import MockFacter
from puppeter import container


def test_fqdn_on_redhat():
    # given
    gateway = FqdnSetterGatewayImpl()
    config = FqdnConfiguration(fqdn='appserver.acme.internal')
    MockFacter.set_fact(Docker, Docker.NO)
    MockFacter.set_fact(OsFamily, OsFamily.RedHat)
    container.initialize()

    # when
    configurers = gateway.process_fully_qualified_domain_name(config)

    # then
    assert len(configurers) == 1

    # when
    configurer = configurers[0]
    commands = configurer.produce_commands()

    # then
    assert u"    puppet resource host 'appserver.acme.internal' " \
        "ensure=present host_aliases='appserver' ip=127.0.0.1 " \
        "comment='FQDN'" in commands
    assert "    hostname 'appserver'" in commands
    assert "  changes => 'set HOSTNAME appserver'" in commands


def test_fqdn_on_debian():
    # given
    gateway = FqdnSetterGatewayImpl()
    config = FqdnConfiguration(fqdn='appserver.acme.internal')
    MockFacter.set_fact(Docker, Docker.NO)
    MockFacter.set_fact(OsFamily, OsFamily.Debian)
    container.initialize()

    # when
    configurers = gateway.process_fully_qualified_domain_name(config)

    # then
    assert len(configurers) == 1

    # when
    configurer = configurers[0]
    commands = configurer.produce_commands()

    # then
    assert u"    puppet resource host 'appserver.acme.internal' " \
           "ensure=present host_aliases='appserver' ip=127.0.0.1 " \
           "comment='FQDN'" in commands
    assert "    hostname 'appserver'" in commands
    assert u"  content => 'appserver'" in commands


def test_fqdn_on_docker():
    # given
    gateway = FqdnSetterGatewayImpl()
    config = FqdnConfiguration(fqdn='appserver.acme.internal')
    MockFacter.set_fact(Docker, Docker.YES)
    container.initialize()

    # when
    with pytest.raises(NotImplementedError) as excinfo:
        gateway.process_fully_qualified_domain_name(config)

    # then
    assert "Can't set FQDN when running inside Docker!" in str(excinfo.value)



import collections
from abc import ABCMeta, abstractmethod

from six import with_metaclass, iteritems
from typing import Dict, Any


class WithOptions(with_metaclass(ABCMeta, object)):

    @abstractmethod
    def raw_options(self):
        # type: () -> Dict[str, Any]
        pass

    @abstractmethod
    def read_raw_options(self, options):
        # type: (Dict[str, Any]) -> None
        pass

    @staticmethod
    def _update(orig_dict, new_dict):
        for key, val in iteritems(new_dict):
            if isinstance(val, collections.Mapping):
                tmp = WithOptions._update(orig_dict=orig_dict.get(key, {}), new_dict=val)
                orig_dict[key] = tmp
            elif isinstance(val, list):
                orig_dict[key] = (orig_dict.get(key, []) + val)
            else:
                orig_dict[key] = new_dict[key]
        return orig_dict



from abc import ABCMeta

from six import with_metaclass

from puppeter.persistence.gateway.installer.debian.base import BaseDebianConfigurer


class AfterPuppet4xConfigurer(with_metaclass(ABCMeta, BaseDebianConfigurer)):
    def __init__(self, installer):
        super(AfterPuppet4xConfigurer, self).__init__(installer)

    def _collect_agent(self, collector):
        collector.collect_from_file('Puppet Agent setup', 'puppetagent.sh')

    def _collect_server(self, collector):
        collector.collect_from_file('Puppet Server setup', 'puppetserver.sh')



import re

from six import iterkeys, itervalues
from typing import Callable, Sequence, MutableSequence

from puppeter.domain import default


class GemVersion:
    VERSION_PATTERN = '[0-9]+(?:\.[0-9a-zA-Z]+)*(-[0-9A-Za-z-]+(\.[0-9A-Za-z-]+)*)?'
    ANCHORED_VERSION_PATTERN = re.compile('^\s*({VERSION_PATTERN})?\s*$'.format(VERSION_PATTERN=VERSION_PATTERN))

    def __init__(self, version):
        # If version is an empty string convert it to 0
        version = 0 if re.compile('^\s*$').match(str(version)) else version

        self.__version = str(version).strip().replace('-', '.pre.')
        self.__segments = None
        self.__bump = None
        self.__release = None

    def bump(self):
        if not self.__bump:
            segments = self.segments()
            while any(map(lambda s: isinstance(s, str), segments)):
                segments.pop()
            if len(segments) > 1:
                segments.pop()
            segments[-1] = segments[-1] + 1
            segments = list(map(lambda r: str(r), segments))
            self.__bump = GemVersion('.'.join(segments))

        return self.__bump

    def release(self):
        if not self.__release:
            segments = self.segments()
            while any(map(lambda s: isinstance(s, str), segments)):
                segments.pop()
            segments = list(map(lambda r: str(r), segments))
            self.__release = GemVersion('.'.join(segments))

        return self.__release

    def segments(self):
        # type: () -> MutableSequence[int|str]
        return list(self.__get_segments())

    def __cmp__(self, other):
        # type: (GemVersion) -> int
        if self.__version == other.__version:
            return 0
        lhsegments = self.__get_segments()
        rhsegments = other.__get_segments()

        lhsize = len(lhsegments)
        rhsize = len(rhsegments)
        limit = (lhsize if lhsize > rhsize else rhsize) - 1

        i = 0

        while i <= limit:
            lhs = default(lambda: lhsegments[i], IndexError, 0)
            rhs = default(lambda: rhsegments[i], IndexError, 0)
            i += 1

            if lhs == rhs:
                continue
            if isinstance(lhs, str) and isinstance(rhs, int):
                return -1
            if isinstance(lhs, int) and isinstance(rhs, str):
                return 1

            return lhs - rhs
        return 0

    def __lt__(self, other):
        return self.__cmp__(other) < 0

    def __le__(self, other):
        return self.__cmp__(other) <= 0

    def __eq__(self, other):
        return self.__cmp__(other) == 0

    def __repr__(self):
        return 'GemVersion({segments})'.format(segments=self.segments())

    def __get_segments(self):
        # type: () -> Sequence[int|str]
        if not self.__segments:
            rex = re.compile('[0-9]+|[a-z]+', re.IGNORECASE)
            d_rex = re.compile('^\d+$')
            self.__segments = tuple(map(lambda s: int(s) if d_rex.match(s) else s, rex.findall(self.__version)))
        return self.__segments


class GemRequirement:
    OPS = {
        '=': lambda v, r: v == r,
        '!=': lambda v, r: v != r,
        '>': lambda v, r: v > r,
        '<': lambda v, r: v < r,
        '>=': lambda v, r: v >= r,
        '<=': lambda v, r: v <= r,
        '~>': lambda v, r: v >= r and v.release() < r.bump()
    }

    PATTERN_RAW = "\\s*({quoted})?\\s*({VERSION_PATTERN})\\s*".format(
        quoted='|'.join(tuple(map(lambda k: re.escape(k), iterkeys(OPS)))),
        VERSION_PATTERN=GemVersion.VERSION_PATTERN
    )

    # A regular expression that matches a requirement
    PATTERN = re.compile('^{PATTERN_RAW}$'.format(PATTERN_RAW=PATTERN_RAW))

    ##
    # The default requirement matches any version

    DEFAULT_REQUIREMENT = tuple(['>=', GemVersion(0)])

    class BadRequirementError(AttributeError):
        pass

    def __init__(self, *requirements):
        # type: (str) -> None
        if len(requirements) == 0:
            self.__requirements = tuple([GemRequirement.DEFAULT_REQUIREMENT])
        else:
            self.__requirements = tuple(map(lambda req: GemRequirement.parse(req), requirements))

    @classmethod
    def parse(cls, requirement):
        if isinstance(requirement, GemVersion):
            return tuple(['=', requirement])

        match = cls.PATTERN.match(str(requirement))
        if not match:
            raise cls.BadRequirementError('Illformed requirement [{inspect}]'.format(inspect=repr(requirement)))

        if match.group(1) == '>=' and match.group(2) == '0':
            return cls.DEFAULT_REQUIREMENT
        else:
            op = match.group(1) if match.group(1) else '='
            return tuple([op, GemVersion(match.group(2))])

    def satified_by(self, version):
        gemver = version if isinstance(version, GemVersion) else GemVersion(version)
        operation = self.__test_rv(gemver)
        return all(map(operation, self.__requirements))

    @classmethod
    def __test_rv(cls, version):
        # type: (GemVersion) -> Callable[[str, GemVersion], bool]
        def __testing(req):
            op, rv = req
            callable = cls.__get_operation(op)
            return callable(version, rv)
        return __testing

    @classmethod
    def __get_operation(cls, op):
        # type: (str) -> Callable[[GemVersion, GemVersion], bool]
        try:
            return cls.OPS[op]
        except KeyError:
            return cls.OPS['=']



from puppeter.container import Named
from puppeter.domain.model.ordered import Order
from puppeter.persistence.gateway.installer.debian.base import BaseDebianConfigurer


@Named('system-debian')
@Order(100)
class DebianSystemConfigurer(BaseDebianConfigurer):
    def _collect_repo(self, collector):
        pass

    def _is_wget_needed(self):
        return False

    def _collect_agent(self, collector):
        collector.collect_from_file('Puppet agent setup', 'puppet-system.sh')

    def _collect_server(self, collector):
        collector.collect_from_file('Puppet Master setup', 'puppetmaster-system.sh')



from puppeter import container
from puppeter.domain.gateway.answers import AnswersProcessor
from puppeter.presentation.answersprocessor import AnswersProcessorImpl
from puppeter.presentation.app import App
from puppeter.presentation.interactiveapp import InteractiveApp
from puppeter.presentation.unattendedapp import UnattendedApp

container.bind(AnswersProcessor, AnswersProcessorImpl)
container.bind(App, InteractiveApp)
container.bind(App, UnattendedApp)



from typing import IO

import puppeter
from puppeter import container
from puppeter.container import Named
from puppeter.domain.gateway.answers import AnswersGateway
from puppeter.domain.model.answers import Answers
from puppeter.presentation.app import App


@Named('unattended')
class UnattendedApp(App):

    def __init__(self, options):
        App.__init__(self, options)
        self.__log = puppeter.get_logger(UnattendedApp)

    def _collect_answers(self):
        target = self._options.answers()
        answers = self.__load_answers(container.get(AnswersGateway), target)
        if self._options.execute():
            self.__log.warning('Installation will be performed from answer file:'
                               ' %s. System will be altered!!!', target.name)
        else:
            self.__log.info('Installation commands will be generated based on answers file'
                            ' %s and printed out. System will NOT be altered.', target.name)
        return answers

    @staticmethod
    def __load_answers(gateway, target):
        # type: (AnswersGateway, IO) -> Answers
        return gateway.read_answers_from_file(target)



from abc import ABCMeta, abstractmethod
from enum import Enum
from typing import Sequence, Any, Dict

from six import with_metaclass

from puppeter import container


class ScriptFormat(Enum):
    BASH = 1
    PUPPET = 2


class CommandsCollector(with_metaclass(ABCMeta)):
    @abstractmethod
    def collect_from_template(self, description, template, mapping, format=ScriptFormat.BASH):
        # type: (str, str, Dict[str, Any]) -> CommandsCollector
        pass

    @abstractmethod
    def collect_from_file(self, description, path):
        # type: (str, str) -> CommandsCollector
        pass

    @abstractmethod
    def lines(self):
        # type: () -> Sequence[str]
        pass


class Configurer(with_metaclass(ABCMeta)):
    @abstractmethod
    def produce_commands(self):
        # type: () -> Sequence[str]
        pass

    def _collector(self):
        # type: () -> CommandsCollector
        impl = container.get_bean(CommandsCollector).impl_cls()
        return impl(self)



from abc import ABCMeta, abstractmethod

from six import with_metaclass
from typing import Sequence
from puppeter.domain.facter import Facter
from puppeter.domain.gateway.fqdn import FqdnSetterGateway
from puppeter.domain.model.configurer import Configurer, CommandsCollector, ScriptFormat
from puppeter.domain.model.fqdn import FqdnConfiguration
from puppeter.domain.model.ordered import Order
from puppeter.domain.model.osfacts import OsFamily, Docker

FQDN_ORDER = 200


class LinuxFQDNConfigurer(with_metaclass(ABCMeta, Configurer)):
    def __init__(self, fqdn):
        self._fqdn = fqdn  # type: FqdnConfiguration

    def produce_commands(self):
        # type: () -> Sequence[str]
        collector = self._collector()
        collector.collect_from_template(
            'Setting FQDN (Fully Qualified Domain Name)',
            'set-fqdn.sh',
            self._mapping()
        )
        self._persist_change(collector)
        return collector.lines()

    @abstractmethod
    def _persist_change(self, collector):
        # type: (CommandsCollector) -> None
        pass

    def _mapping(self):
        return dict(fqdn=self._fqdn.fqdn(), hostname=self._fqdn.hostname())


@Order(FQDN_ORDER)
class RedHatFQDNConfigurer(LinuxFQDNConfigurer):
    def _persist_change(self, collector):
        collector.collect_from_template(
            'Persist FQDN to be available after reboot',
            'redhat-persist-fqdn.pp',
            self._mapping(),
            format=ScriptFormat.PUPPET
        )


@Order(FQDN_ORDER)
class DebianFQDNConfigurer(LinuxFQDNConfigurer):
    def _persist_change(self, collector):
        collector.collect_from_template(
            'Persist FQDN to be available after reboot',
            'debian-persist-fqdn.pp',
            self._mapping(),
            format=ScriptFormat.PUPPET
        )


class FqdnSetterGatewayImpl(FqdnSetterGateway):
    def process_fully_qualified_domain_name(self, fqdn):
        # type: (FqdnConfiguration) -> Sequence[Configurer]
        if Facter.get(Docker) == Docker.YES:
            raise NotImplementedError('Can\'t set FQDN when running inside Docker!')
        osfamily = Facter.get(OsFamily)
        try:
            return {
                OsFamily.RedHat: [RedHatFQDNConfigurer(fqdn)],
                OsFamily.Debian: [DebianFQDNConfigurer(fqdn)]
            }[osfamily]
        except KeyError as ex:
            raise NotImplementedError(ex, 'Unsupported OS family %s for setting FQDN' % osfamily)



import argparse
import sys

import puppeter
from puppeter import container
from puppeter.presentation import App
from puppeter.presentation.app import Options


class _VersionAction(argparse.Action):

    def __init__(self,
                 option_strings,
                 version=None,
                 dest=argparse.SUPPRESS,
                 default=argparse.SUPPRESS,
                 help="show program's version number and exit"):
        super(_VersionAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)
        self.version = version

    def __call__(self, parser, namespace, values, option_string=None):
        version = self.version
        if version is None:
            version = parser.version
        formatter = parser._get_formatter()
        formatter.add_text(version)
        parser._print_message(formatter.format_help(), sys.stdout)
        parser.exit()


class StdErrArgumentParser(argparse.ArgumentParser):
    def print_help(self, file=None):
        if file is None:
            file = sys.stderr
        self._print_message(self.format_help(), file)


class CommandLineParser(object):
    """CommandLineParser for Puppeter"""
    def __init__(self, argv):
        super(CommandLineParser, self).__init__()
        self.__argv = argv[1:]

    def parse(self):
        # type: () -> App
        parser = StdErrArgumentParser(prog='puppeter', description='Puppeter - an automatic puppet installer',
                                      epilog='By default interactive setup is performed and chosen values can be saved'
                                      ' to answer file.')
        parser.add_argument('--answers', '-a', type=argparse.FileType('r'),
                            metavar='FILE',
                            help='An answer file to be used to perform unattended setup')
        parser.add_argument('--verbose', '-v', action='count',
                            help='Print more verbose output (up to 2 verbosity flags are supported)')
        parser.add_argument('--version', action=_VersionAction, version='%(prog)s ' + puppeter.__version__)
        parser.add_argument('--execute', '-e', action='store_true',
                            help='Executes setup commands instead of printing them')

        parsed = parser.parse_args(self.__argv)
        options = Options(parsed)
        apptype = 'interactive' if options.answers() is None else 'unattended'
        return container.get_named(App, apptype, options=options)



from puppeter import container
from puppeter.domain.model import Collection3xInstaller
from puppeter.domain.model.osfacts import OperatingSystem, OsFamily, OperatingSystemRelease
from puppeter.persistence.gateway.installer.redhat import RedHatPC3xConfigurer
from tests.domain.mock_facter import MockFacter


def test_pc3x_centos_5():
    # given
    MockFacter.set_fact(OperatingSystem, OperatingSystem.CentOS)
    MockFacter.set_fact(OperatingSystemRelease, OperatingSystemRelease('5.4'))
    installer = Collection3xInstaller()
    installer.read_raw_options({'mode': 'Server'})
    configurer = RedHatPC3xConfigurer(installer=installer)

    # when
    commands = configurer.produce_commands()
    commands = list(map(lambda cmd: cmd.strip(), commands))

    # then
    assert 'yum install -y wget' in commands
    assert "wget 'https://yum.puppetlabs.com/puppetlabs-release-el-5.noarch.rpm'" in commands
    assert 'yum install -y puppet' in commands
    assert 'puppet resource package puppetmaster ensure=installed' in commands


def test_pc3x_scientific_7():
    # given
    MockFacter.set_fact(OperatingSystem, OperatingSystem.Scientific)
    MockFacter.set_fact(OperatingSystemRelease, OperatingSystemRelease('7.3'))
    installer = Collection3xInstaller()
    installer.read_raw_options({'mode': 'Server'})
    configurer = RedHatPC3xConfigurer(installer=installer)

    # when
    commands = configurer.produce_commands()
    commands = list(map(lambda cmd: cmd.strip(), commands))

    # then
    assert 'yum install -y wget' not in commands
    assert "rpm -Uvh 'https://yum.puppetlabs.com/puppetlabs-release-el-7.noarch.rpm'" in commands
    assert 'yum install -y puppet' in commands
    assert 'puppet resource package puppetmaster ensure=installed' in commands


def test_pc3x_fedora_25():
    # given
    MockFacter.set_fact(OperatingSystem, OperatingSystem.Fedora)
    MockFacter.set_fact(OperatingSystemRelease, OperatingSystemRelease('25.1'))
    installer = Collection3xInstaller()
    installer.read_raw_options({'mode': 'Agent'})
    configurer = RedHatPC3xConfigurer(installer=installer)

    # when
    commands = configurer.produce_commands()
    commands = list(map(lambda cmd: cmd.strip(), commands))

    # then
    assert 'yum install -y wget' not in commands
    assert "rpm -Uvh 'https://yum.puppetlabs.com/puppetlabs-release-fedora-25.noarch.rpm'" in commands
    assert 'yum install -y puppet' in commands
    assert 'puppet resource package puppetmaster ensure=installed' not in commands


def setup_function():
    MockFacter.set_fact(OsFamily, OsFamily.RedHat)
    container.initialize()


def teardown_function():
    MockFacter.reset_facts()



from enum import Enum


class JavaVersion(Enum):
    NOT_INSTALLED = 1
    JRE6 = 2
    JRE7 = 3
    JRE8 = 4
    JRE9 = 5

    def has_permgen_space(self):
        return self in (JavaVersion.JRE7, JavaVersion.JRE7)



from puppeter import container
from puppeter.domain.model import Collection4xInstaller
from puppeter.domain.model.osfacts import OsFamily, OperatingSystem, OperatingSystemRelease
from puppeter.persistence.gateway.installer.redhat import RedHatPC4xConfigurer
from tests.domain.mock_facter import MockFacter


def test_pc4x_oraclelinux5():
    # given
    MockFacter.set_fact(OperatingSystem, OperatingSystem.OracleLinux)
    MockFacter.set_fact(OperatingSystemRelease, OperatingSystemRelease('5.11'))
    installer = Collection4xInstaller()
    installer.read_raw_options({'mode': 'Server'})
    configurer = RedHatPC4xConfigurer(installer=installer)

    # when
    commands = configurer.produce_commands()
    commands = list(map(lambda cmd: cmd.strip(), commands))

    # then
    assert 'yum install -y wget' in commands
    assert "wget 'https://yum.puppetlabs.com/puppetlabs-release-pc1-el-5.noarch.rpm'" in commands
    assert 'yum install -y puppet-agent' in commands
    assert 'puppet resource package puppetserver ensure=installed' in commands


def test_pc4x_rhel_6():
    # given
    MockFacter.set_fact(OperatingSystem, OperatingSystem.RedHat)
    MockFacter.set_fact(OperatingSystemRelease, OperatingSystemRelease('6.8'))
    installer = Collection4xInstaller()
    installer.read_raw_options({'mode': 'Agent'})
    configurer = RedHatPC4xConfigurer(installer=installer)

    # when
    commands = configurer.produce_commands()
    commands = list(map(lambda cmd: cmd.strip(), commands))

    # then
    assert 'yum install -y wget' not in commands
    assert "rpm -Uvh 'https://yum.puppetlabs.com/puppetlabs-release-pc1-el-6.noarch.rpm'" in commands
    assert 'yum install -y puppet-agent' in commands
    assert 'puppet resource package puppetserver ensure=installed' not in commands


def setup_function():
    MockFacter.set_fact(OsFamily, OsFamily.RedHat)
    container.initialize()


def teardown_function():
    MockFacter.reset_facts()



from abc import ABCMeta, abstractmethod

from six import with_metaclass
from typing import Sequence

from puppeter.domain.model.configurer import Configurer


class ScriptPostProcessor(with_metaclass(ABCMeta)):
    @abstractmethod
    def postprocess(self, commands):
        # type: (Sequence[str]) -> Sequence[str]
        pass


class ScriptLibrariesConfigurer(with_metaclass(ABCMeta, Configurer)):
    pass



import collections

from six import iteritems
from typing import Sequence, Tuple

from puppeter.container import Named
from puppeter.domain.model.configurer import Configurer, ScriptFormat
from puppeter.domain.model.installer import PuppetConf
from puppeter.domain.model.ordered import Order


@Named('puppet.conf')
@Order(900)
class PuppetConfConfigurer(Configurer):

    def __init__(self, installer):
        self.__puppetconf = installer.puppetconf()  # type: PuppetConf

    def produce_commands(self):
        # type: () -> Sequence[str]
        collector = self._collector()
        augeas = self.__puppetconf_as_augeas()
        augeas_pp = ',\n    '.join(list(map(lambda cmd: '"{cmd}"'.format(cmd=cmd), augeas)))
        collector.collect_from_template(
            description="Configuring Puppet configuration file (puppet.conf)",
            format=ScriptFormat.PUPPET,
            template='puppetconf.pp',
            mapping={
                'settings': augeas_pp
            }
        )
        return collector.lines()

    def __puppetconf_as_augeas(self):
        # type: () -> Sequence[str]
        flat = self.__flatten(self.__puppetconf)
        return list(map(PuppetConfConfigurer.__augeaize, iteritems(flat)))

    @staticmethod
    def __augeaize(tup):
        # type: (Tuple[str, str]) -> str
        key, value = tup
        try:
            safe = {
                str: lambda x: "'{str}'".format(str=x),
                bool: lambda x: 'true' if x else 'false',
            }[value.__class__](value)
        except KeyError:
            safe = value
        return "set {key} {value}".format(key=key, value=safe)

    @staticmethod
    def __flatten(d, parent_key='', sep='/'):
        items = []
        for k, v in iteritems(d):
            new_key = parent_key + sep + k if parent_key else k
            if isinstance(v, collections.MutableMapping):
                items.extend(PuppetConfConfigurer.__flatten(v, new_key, sep=sep).items())
            else:
                items.append((new_key, v))
        return dict(items)



from puppeter import container
from puppeter.domain.model.installer import Collection4xInstaller,\
    Collection3xInstaller,\
    Collection5xInstaller,\
    Installer,\
    RubygemsInstaller, \
    SystemInstaller

# TODO: implement GEM installer - https://github.com/coi-gov-pl/puppeter/issues/18
# container.bind(Installer, RubygemsInstaller)
container.bind(Installer, SystemInstaller)
container.bind(Installer, Collection3xInstaller)
container.bind(Installer, Collection4xInstaller)
container.bind(Installer, Collection5xInstaller)



from abc import ABCMeta, abstractmethod

from six import with_metaclass

from puppeter.domain.model import Installer
from puppeter.domain.model.configurer import Configurer, CommandsCollector
from puppeter.domain.model.installer import Mode


class BaseDebianConfigurer(with_metaclass(ABCMeta, Configurer)):
    UPDATE_INTERVAL = 24 * 60 * 60

    def __init__(self, installer):
        self._installer = installer  # type: Installer

    def produce_commands(self):
        collector = self._collector()
        collector.collect_from_template('apt system update',
                                        'apt-update.sh',
                                        dict(interval=self.UPDATE_INTERVAL))
        if self._is_wget_needed():
            collector.collect_from_file('wget downloader', 'wget.sh')
        self._collect_repo(collector)
        self._collect_agent(collector)
        if self._installer.mode() == Mode.Server:
            self._collect_server(collector)
        return collector.lines()

    def _is_wget_needed(self):
        return True

    @abstractmethod
    def _collect_repo(self, collector):
        # type: (CommandsCollector) -> None
        raise NotImplementedError()

    @abstractmethod
    def _collect_agent(self, collector):
        # type: (CommandsCollector) -> None
        raise NotImplementedError()

    @abstractmethod
    def _collect_server(self, collector):
        # type: (CommandsCollector) -> None
        raise NotImplementedError()



import sys

from puppeter import container
from puppeter.presentation.cmdparser import CommandLineParser


def main(argv=sys.argv):
    """Entry point for the puppeter application"""
    container.initialize()
    parser = CommandLineParser(argv)
    app = parser.parse()
    app.run()



from __future__ import absolute_import
import re
import subprocess

from os import path, environ, access, X_OK

from puppeter.domain.facter import Facter
from puppeter.domain.model.javafacts import JavaVersion
from puppeter.domain.model.osfacts import OsFamily


def calculate_java_version():
    # type: () -> JavaVersion
    osfamily = Facter.get(OsFamily)
    java_prog_name = 'java.exe' if osfamily is OsFamily.Windows else 'java'
    search_paths = environ['PATH'].split(path.pathsep)
    java_prog = None
    for search_path in search_paths:
        java_prog_candidate = path.join(search_path, java_prog_name)
        if __is_executable(java_prog_candidate):
            java_prog = java_prog_candidate
            break
    if java_prog is None:
        return JavaVersion.NOT_INSTALLED
    return __get_java_version(java_prog)


def __is_executable(candidate):
    return path.exists(candidate) and access(candidate, X_OK)


def __get_java_version(java_prog):
    sp = subprocess.Popen([java_prog, '-version'],
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE,
                          universal_newlines=True)
    (x, stderr) = sp.communicate()
    retcode = sp.wait()
    if retcode != 0:
        return JavaVersion.NOT_INSTALLED
    pattern = re.compile('(?:jdk|jre|java) version "([0-9]+)\.([0-9]+)\.([0-9]+)[_-]([0-9a-z_-]+)"')
    for line in stderr.splitlines():
        match = pattern.match(line)
        if match is not None:
            ver = match.group(2)
            return {
                '6': JavaVersion.JRE6,
                '7': JavaVersion.JRE7,
                '8': JavaVersion.JRE8,
                '9': JavaVersion.JRE9
            }[ver]
    return JavaVersion.NOT_INSTALLED



from typing import Union

import pytest

from puppeter.container import NamedBean
from puppeter.domain.model.installer import WithOptions, After4xCollectionInstaller


@pytest.mark.skip('not yet implemented')
def test_rubygems_installer_named_bean():
    # given
    from puppeter.domain.model.installer import RubygemsInstaller
    installer = RubygemsInstaller()
    # when
    bean_name = installer.bean_name()
    # then
    assert bean_name is 'gem'


@pytest.mark.skip('not yet implemented')
def test_rubygems_installer_values():
    # given
    from puppeter.domain.model.installer import RubygemsInstaller, Mode
    installer = RubygemsInstaller()
    # when
    version = installer.version()
    mode = installer.mode()
    # then
    assert version is None
    assert mode is Mode.Agent


def test_rubygems_installer_raw_options():
    # given
    from puppeter.domain.model.installer import RubygemsInstaller, Installer
    installer = RubygemsInstaller()
    # hack to test abs method
    Installer.is_after_4x(installer)
    # when
    options = installer.raw_options()
    # then
    assert options == {
        'mode': 'Agent',
        'version': None,
        'type': 'gem'
    }


def test_rubygems_installer_read_raw_options():
    # given
    from puppeter.domain.model.installer import RubygemsInstaller, Mode
    installer = RubygemsInstaller()
    # when
    installer.read_raw_options({
        'mode': 'Server',
        'version': '~> 3'
    })
    version = installer.version()
    mode = installer.mode()
    # then
    assert version is '~> 3'
    assert mode is Mode.Server
    assert installer.is_after_4x() is False


def test_rubygems_installer_read_raw_options_invalid():
    # given
    from puppeter.domain.model.installer import RubygemsInstaller, Mode
    installer = RubygemsInstaller()
    # when
    installer.read_raw_options({})
    installer.read_raw_options({
        'invalid-key': 'Server',
        'version': '~> 4'
    })
    mode = installer.mode()

    # then
    assert mode is Mode.Agent
    assert installer.is_after_4x()


def test_system_installer_is_after_4x():
    # given
    from puppeter.domain.model.installer import Installer, Mode
    from puppeter import container

    # when
    installer = container.get_named(Installer, 'system')  # type: Union[Installer, NamedBean]

    # then
    assert installer.bean_name() is 'system'
    assert installer.is_after_4x() is False


def test_getting_impl_from_container():
    # given
    from puppeter.domain.model.installer import Installer, Mode
    from puppeter import container

    # when
    installer = container.get_named(Installer, 'pc3x')  # type: Union[Installer, NamedBean]

    # then
    assert installer.mode() is Mode.Agent
    assert installer.bean_name() is 'pc3x'


def test_getting_all_impls_from_container():
    # given
    from puppeter.domain.model.installer import Installer, Collection5xInstaller
    from puppeter import container

    # when
    installers = container.get_all(Installer)

    # then
    assert len(installers) >= 4
    assert len(tuple(filter(lambda cls: isinstance(cls, Collection5xInstaller), installers))) == 1


def test_with_options():
    # given
    class TestA(WithOptions):
        def read_raw_options(self, options):
            WithOptions.read_raw_options(self, options)

        def raw_options(self):
            return WithOptions.raw_options(self)

    # when
    testa = TestA()

    # then
    assert testa is not None
    assert testa.raw_options() is None
    assert testa.read_raw_options({}) is None


def test_pc4x_jvmargs_read():
    # given
    from puppeter.domain.model.installer import Installer
    from puppeter import container

    # when
    installer = container.get_named(Installer, 'pc4x')  # type: After4xCollectionInstaller
    installer.read_raw_options({
        'mode': 'Server',
        'puppetserver': {
            'jvm': {
                'memory': {
                    'heap': {
                        'min': '256m',
                        'max': '512m'
                    }
                },
                'args': [
                    '-XX:+UseConcMarkSweepGC',
                    '-XX:+CMSParallelRemarkEnabled'
                ]
            }
        }
    })

    # then
    assert len(installer.puppetserver_jvm_args()) is 2
    assert '-XX:+UseConcMarkSweepGC' in installer.puppetserver_jvm_args()
    assert '-XX:+CMSParallelRemarkEnabled' in installer.puppetserver_jvm_args()
    assert installer.puppetserver_jvm_memory().heap_maximum() == '512m'
    assert installer.puppetserver_jvm_memory().heap_minimum() == '256m'


def test_pc4x_jvmargs_write():
    # given
    from puppeter.domain.model.installer import Installer
    from puppeter import container
    installer = container.get_named(Installer, 'pc4x')  # type: After4xCollectionInstaller

    # when
    installer.read_raw_options({
        'mode': 'Server',
        'puppetserver': {
            'jvm': {
                'memory': {
                    'heap': {
                        'max': '434m'
                    },
                    'metaspace': {
                        'max': '11k'
                    }
                },
                'args': [
                    '-XX:+UseConcMarkSweepGC'
                ]
            }
        }
    })
    raw = installer.raw_options()

    # then
    assert raw['puppetserver']['jvm']['memory']['metaspace']['max'] == '11k'
    assert installer.puppetserver_jvm_memory().metaspace_maximum() == '11k'
    assert installer.puppetserver_jvm_memory().heap_minimum() == '434m'
    assert installer.puppetserver_jvm_memory().heap_maximum() == '434m'
    assert len(installer.puppetserver_jvm_args()) is 1
    assert '-XX:+UseConcMarkSweepGC' in installer.puppetserver_jvm_args()


def test_pc3x_is_after4x():
    # given
    from puppeter.domain.model.installer import Installer
    from puppeter import container
    installer = container.get_named(Installer, 'pc3x')

    # then
    assert installer.is_after_4x() is False


def test_pc4x_puppetconf_read():
    # given
    from puppeter.domain.model.installer import Installer
    from puppeter import container
    installer = container.get_named(Installer, 'pc4x')  # type: After4xCollectionInstaller

    # when
    installer.read_raw_options({
        'mode': 'Agent',
        'puppet.conf': {
            'main': {
                'server': 'puppet.acme.internal',
                'noop': True
            }
        }
    })
    raw = installer.raw_options()

    # then
    assert raw['puppet.conf']['main']['server'] == 'puppet.acme.internal'
    assert installer.puppetconf()['main']['server'] == 'puppet.acme.internal'
    assert installer.puppetconf()['main']['noop'] is True


def test_javamemoryspec_scale():
    # given
    from puppeter.domain.model.installer import JavaMemorySpec, MemScale
    spec = JavaMemorySpec(1024, scale=MemScale.KILO)

    # when
    s = spec.scale()

    # then
    assert s == MemScale.KILO


def test_javamemoryspec_number():
    # given
    from puppeter.domain.model.installer import JavaMemorySpec
    spec = JavaMemorySpec(2048)

    # when
    n = spec.number()

    # then
    assert n == 2048



from enum import Enum

import re


class OsFamily(Enum):
    Unknown = 1
    RedHat = 2
    Debian = 3
    Suse = 4
    Windows = 5
    AIX = 6
    Solaris = 7
    MacOS = 8
    Arch = 9

    def is_posix(self):
        return self not in (OsFamily.Unknown, OsFamily.Windows)


class OperatingSystem(Enum):
    Unknown = 1
    RedHat = 2
    CentOS = 3
    Scientific = 4
    OracleLinux = 5
    Debian = 6
    Ubuntu = 7
    OpenSuse = 8
    Fedora = 9
    SLES = 10
    ArchLinux = 11


class OperatingSystemRelease(str):
    # https://regex101.com/r/uJ8oTf/2
    VERSION_RE = re.compile('^(\d+)(?:\.(\d+))?(?:\.(\d+))?(?:[._\-](.*))?$')

    def __init__(self, version):
        super(OperatingSystemRelease, self).__init__()
        self.__version = str(version)

    def __str__(self):
        return self.__version

    def major(self):
        return self.__group(1)

    def minor(self):
        return self.__group(2)

    def patch(self):
        return self.__group(3)

    def lesser(self):
        return self.__group(4)

    def __group(self, num):
        try:
            return self.VERSION_RE \
                .match(self) \
                .group(num)
        except AttributeError:
            return str(self)


class OperatingSystemCodename(str):
    def __init__(self, codename):
        super(OperatingSystemCodename, self).__init__()
        self.__codename = str(codename)

    def __str__(self):
        return self.__codename


class Docker(Enum):
    NO = 1
    YES = 2

    def __nonzero__(self):
        return self is Docker.YES



from puppeter.container import Named
from puppeter.domain.facter import Facter
from puppeter.domain.model.ordered import Order
from puppeter.domain.model.osfacts import OperatingSystemRelease, OperatingSystem
from puppeter.persistence.gateway.installer.redhat.base import BaseRedHatConfigurer


@Named('pc3x-redhat')
@Order(100)
class RedHatPC3xConfigurer(BaseRedHatConfigurer):

    def __init__(self, installer):
        BaseRedHatConfigurer.__init__(self, installer)

    def _collect_repo(self, collector):
        major = Facter.get(OperatingSystemRelease).major()
        os = Facter.get(OperatingSystem)
        if os == OperatingSystem.Fedora:
            abbr = 'fedora'
        else:
            abbr = 'el'
        mapping = dict(major=major, abbr=abbr)
        if ('%s-%s' % (abbr, major)) == 'el-5':
            script = 'pc3x-el5-repo.sh'
        else:
            script = 'pc3x-repo.sh'
        title = 'Puppet Package Repository setup (Puppet OSS 3.x)'
        collector.collect_from_template(title, script, mapping)

    def _collect_agent(self, collector):
        collector.collect_from_file('Puppet agent setup', 'puppet3xagent.sh')

    def _collect_server(self, collector):
        collector.collect_from_file('PuppetMaster setup', 'puppetmaster.sh')



import re

from puppeter import container
from puppeter.domain.model.osfacts import OperatingSystemCodename,\
    OperatingSystem, OsFamily, OperatingSystemRelease
from puppeter.presentation.app import Options
from puppeter.presentation.unattendedapp import UnattendedApp
from tests.domain.mock_facter import MockFacter


def test_unattendedapp_onlyinstaller_with_verbosity(tmpdir, capsys):
    # given
    tmp = tmpdir.join("answers.yml")
    tmp.write('installer:\n'
              '    type: pc4x\n'
              '    mode: Server\n')
    answers = open(str(tmp))
    options = Options(DotDict(answers=answers, verbose=2, execute=False))
    app = UnattendedApp(options)

    # when
    app.run()
    out, err = capsys.readouterr()

    # then
    assert "wget 'https://apt.puppetlabs.com/puppetlabs-release-pc1-trusty.deb'" in out
    assert 'DEBUG: Answers loaded from file' in err
    assert 'INFO: Installation commands will be generated based on answers file' in err


class DotDict(dict):
    def __getattr__(self, name):
        return self[name]


def restr(st):
    return re.compile('\s*%s' % re.escape(st), re.MULTILINE)


def setup_function():
    container.initialize()
    MockFacter.set_fact(OperatingSystemRelease, OperatingSystemRelease('14.04'))
    MockFacter.set_fact(OperatingSystemCodename, OperatingSystemCodename('trusty'))
    MockFacter.set_fact(OperatingSystem, OperatingSystem.Ubuntu)
    MockFacter.set_fact(OsFamily, OsFamily.Debian)


def teardown_function():
    MockFacter.reset_facts()



from typing import Optional

from puppeter.domain.model.csr import CsrAttributesConfiguration
from puppeter.domain.model.fqdn import FqdnConfiguration
from puppeter.domain.model.installer import Installer


class Answers:
    def __init__(self,
                 installer=None,
                 fqdn_configuration=None,
                 timesync_configuration=None,
                 csrattrs_configuration=None):
        self.__installer = installer  # type: Optional[Installer]
        self.__fqdn_configuration = fqdn_configuration  # type: Optional[FqdnConfiguration]
        self.__timesync_configuration = timesync_configuration
        self.__csrattrs_configuration = csrattrs_configuration  # type: Optional[CsrAttributesConfiguration]

    def installer(self):
        # type: () -> Optional[Installer]
        return self.__installer

    def fqdn_configuration(self):
        # type: () -> Optional[FqdnConfiguration]
        return self.__fqdn_configuration

    def timesync_configuration(self):
        return self.__timesync_configuration

    def csrattrs_configuration(self):
        # type: () -> Optional[CsrAttributesConfiguration]
        return self.__csrattrs_configuration



from abc import ABCMeta, abstractmethod

from six import with_metaclass
from typing import Sequence

from puppeter.domain.model.configurer import Configurer
from puppeter.domain.model.fqdn import FqdnConfiguration


class FqdnSetterGateway(with_metaclass(ABCMeta)):
    @abstractmethod
    def process_fully_qualified_domain_name(self, fqdn):
        # type: (FqdnConfiguration) -> Sequence[Configurer]
        pass



from typing import Sequence, Optional

from puppeter.domain.model import Installer
from puppeter.domain.model.configurer import Configurer


class InstallerGateway:
    def configurers(self, installer: Optional[Installer]) -> Sequence[Configurer]: ...
    def _provide_install_configurers(self, installer: Installer) -> Sequence[Configurer]: ...
    def _puppet_cert_issue(self, installer: Installer) -> Sequence[Configurer]: ...
    def _puppet_services(self, installer: Installer) -> Sequence[Configurer]: ...
    def _puppet_conf(self, installer: Installer) -> Sequence[Configurer]: ...




from puppeter import container
from puppeter.domain.gateway.answers import AnswersGateway
from puppeter.domain.gateway.csr import CsrAttributesSetterGateway
from puppeter.domain.gateway.fqdn import FqdnSetterGateway
from puppeter.domain.gateway.script import ScriptPostProcessor, ScriptLibrariesConfigurer
from puppeter.persistence.gateway.answers import YamlAnswersGateway
from puppeter.persistence.gateway.csr import CsrAttributesSetterGatewayImpl
from puppeter.persistence.gateway.fqdn import FqdnSetterGatewayImpl
from puppeter.persistence.gateway.script import BashScriptPostProcessor, BashScriptLibrariesConfigurer

container.bind(AnswersGateway, YamlAnswersGateway)
container.bind(FqdnSetterGateway, FqdnSetterGatewayImpl)
container.bind(ScriptPostProcessor, BashScriptPostProcessor)
container.bind(ScriptLibrariesConfigurer, BashScriptLibrariesConfigurer)
container.bind(CsrAttributesSetterGateway, CsrAttributesSetterGatewayImpl)



from puppeter.container import Named
from puppeter.domain.facter import Facter
from puppeter.domain.model.ordered import Order
from puppeter.domain.model.osfacts import OperatingSystemCodename
from puppeter.persistence.gateway.installer.debian.base import BaseDebianConfigurer


@Named('pc3x-debian')
@Order(100)
class DebianPC3xConfigurer(BaseDebianConfigurer):
    def __init__(self, installer):
        BaseDebianConfigurer.__init__(self, installer)

    def _collect_repo(self, collector):
        codename = Facter.get(OperatingSystemCodename)
        collector.collect_from_template(
            'Puppet Package Repository setup (Puppet OSS 3.x)',
            'pc3x-repo.sh',
            dict(codename=codename)
        )

    def _collect_agent(self, collector):
        collector.collect_from_file('Puppet agent setup', 'puppet3xagent.sh')

    def _collect_server(self, collector):
        collector.collect_from_file('PuppetMaster setup', 'puppetmaster.sh')



from puppeter.container import Named
from puppeter.domain.model.ordered import Order
from puppeter.persistence.gateway.installer.redhat.after4x import AfterPuppet4xConfigurer


@Named('pc5x-redhat')
@Order(100)
class RedHatPC5xConfigurer(AfterPuppet4xConfigurer):

    def __init__(self, installer):
        AfterPuppet4xConfigurer.__init__(self, installer)

    def _repo_script_path_el5(self):
        return 'pc5x-el5-repo.sh'

    def _repo_script_path(self):
        return 'pc5x-repo.sh'

    def _repo_setup_title(self):
        return 'Puppet Platform Repository setup (Puppet OSS 5.x)'



def Order(order):
    # type: (int) -> callable
    def order_decorator(cls):

        class Ordered(cls):
            def __init__(self, *args, **kwargs):
                self.wrapped = cls(*args, **kwargs)

            @staticmethod
            def order():
                return order

            @staticmethod
            def original_cls():
                return cls

            def __repr__(self):
                return '@Order(\'%d\') %s' % (order, repr(self.wrapped))

            def __getattr__(self, name):
                return getattr(self.wrapped, name)
        return Ordered
    return order_decorator



from abc import ABCMeta, abstractmethod

from six import with_metaclass
from os.path import dirname
from typing import Generic, TypeVar, cast, Any, Callable, Union, Type, \
    Sequence, MutableSequence, Dict, Optional

from puppeter import __program__

T = TypeVar('T')


def initialize():
    __load_modules(__program__)


class NamedBean(with_metaclass(ABCMeta)):
    @staticmethod
    @abstractmethod
    def bean_name():
        # type: () -> str
        pass

    @staticmethod
    @abstractmethod
    def original_cls():
        # type: () -> Type
        pass


def Named(bean_name):
    def named_decorator(cls):

        class NamedBeanImpl(NamedBean, cls):
            def __init__(self, *args, **kwargs):
                self.wrapped = cls(*args, **kwargs)

            @staticmethod
            def bean_name():
                return bean_name

            @staticmethod
            def original_cls():
                return cls

            def __repr__(self):
                return '@Named(\'%s\') %s' % (bean_name, repr(self.wrapped))

            def __getattr__(self, name):
                return getattr(self.wrapped, name)
        return NamedBeanImpl
    return named_decorator


class Bean(Generic[T]):
    def __init__(self, cls, impl=None, impl_cls=None):
        if impl is None and impl_cls is None:
            raise Exception('20170707:164636')
        self.__cls = cls  # type: Type[T]
        self.__impl = impl  # type: T
        self.__impl_cls = impl_cls  # type: Union[Type[T], Callable[[Any, Any], T], NamedBean]

    def __repr__(self):
        name = self.name()
        if name is not None:
            return 'Bean named \'%s\' of %s' % (name, repr(self.impl_cls()))
        else:
            return 'Bean of %s' % repr(self.impl_cls())

    def impl_cls(self):
        # type: () -> Union[Type[T], Callable[[Any, Any], T], NamedBean]
        if self.__impl is None and self.__impl_cls is not None:
            return self.__impl_cls
        else:
            return self.__impl.__class__

    def name(self):
        # type: () -> Optional[str]
        try:
            impl_cls = cast(NamedBean, self.impl_cls())
            return impl_cls.bean_name()
        except AttributeError:
            return None

    def impl(self, *args, **kwargs):
        # type: (Any, Any) -> T
        if self.__impl is None and self.__impl_cls is not None:
            self.__impl = self.__impl_cls(*args, **kwargs)
        # noinspection PyTypeChecker
        return self.__impl


class Container:

    def __init__(self):
        self.__beans = {}  # type: Dict[Type, MutableSequence[Bean]]

    def bind(self, cls, impl_cls):
        # type: (Type[T], Type[T]) -> None
        beans = self.__get_all_beans(cls)
        beans.append(Bean(cls, impl_cls=impl_cls))

    def get_all(self, cls, *args, **kwargs):
        # type: (Type[T], Any, Any) -> Sequence[T]
        beans = self.__get_all_beans(cls)
        try:
            return self.__instantinate_beans(beans, *args, **kwargs)
        except Exception:
            return tuple()

    def get(self, cls, *args, **kwargs):
        # type: (Type[T], Any, Any) -> T
        beans = self.__get_all_beans(cls)
        if len(beans) == 1:
            return beans[0].impl(*args, **kwargs)
        else:
            impls = list(map(lambda bean: str(bean.impl_cls()), beans))
            raise ValueError('Zero or more then one implementation found for class %s. '
                             'Found those implementations: %s. '
                             'Use @Named beans and get_named() function!' % (cls, impls))

    def get_named(self, cls, bean_name, *args, **kwargs):
        # type: (Type[T], str, Any, Any) -> T
        for bean in self.__get_all_beans(cls):
            if bean.name() == bean_name:
                return bean.impl(*args, **kwargs)
        raise ValueError('Bean named \'%s\' has not been found for class %s' % (bean_name, cls))

    def get_all_with_name_starting_with(self, cls, name_prefix, *args, **kwargs):
        # type: (Type[T], str, Any, Any) -> Sequence[T]
        beans = []
        for bean in self.__get_all_beans(cls):
            name = bean.name()
            if name is not None and name.startswith(name_prefix):
                beans.append(bean)
        return self.__instantinate_beans(beans, *args, **kwargs)

    def get_bean(self, cls):
        # type: (Type[T]) -> Bean[T]
        return self.__get_all_beans(cls)[0]

    def __instantinate_beans(self, beans, *args, **kwargs):
        # type: (Sequence[Bean[T]], Any, Any) -> Sequence[T]
        impls = tuple(map(lambda bean: bean.impl(*args, **kwargs), beans))  # type: Sequence[T]
        return impls

    def __get_all_beans(self, cls):
        # type: (Type[T]) -> MutableSequence[Bean[T]]
        try:
            return self.__beans[cls]
        except KeyError:
            lst = []  # type: MutableSequence[Bean[T]]
            self.__beans[cls] = lst
            return lst


app_container = Container()


def bind(cls, impl_cls):
    # type: (Type[T], Type[T]) -> None
    app_container.bind(cls, impl_cls)


def get_all(cls, *args, **kwargs):
    # type: (Type[T], Any, Any) -> Sequence[T]
    return app_container.get_all(cls, *args, **kwargs)


def get(cls, *args, **kwargs):
    # type: (Type[T], Any, Any) -> T
    return app_container.get(cls, *args, **kwargs)


def get_named(cls, bean_name, *args, **kwargs):
    # type: (Type[T], str, Any, Any) -> T
    return app_container.get_named(cls, bean_name, *args, **kwargs)


def get_all_with_name_starting_with(cls, name_prefix, *args, **kwargs):
    # type: (Type[T], str, Any, Any) -> Sequence[T]
    return app_container.get_all_with_name_starting_with(cls, name_prefix, *args, **kwargs)


def get_bean(cls):
    return app_container.get_bean(cls)


def __load_modules(module_name):
    import os
    from os.path import join, abspath, isdir, exists
    rootdir = dirname(dirname(__file__))

    search = join(abspath(rootdir), module_name.replace('.', os.sep))
    lst = os.listdir(search)
    modules = []
    for d in lst:
        subpath = join(search, d)
        if isdir(subpath) and exists(join(subpath, '__init__.py')):
            submodule_name = module_name + '.' + d
            __load_modules(submodule_name)
            modules.append(submodule_name)
    # load the modules
    for module_name_to_import in modules:
        __import__(module_name_to_import)



import pytest

from integration_tests.phial import Phial, \
    ssh_service, \
    docker_ip, \
    docker_services, \
    docker_compose_file, \
    docker_compose_project_name, \
    docker_allow_fallback

__fixtures = [
    ssh_service,
    docker_ip,
    docker_services,
    docker_compose_file,
    docker_compose_project_name,
    docker_allow_fallback
]


@pytest.fixture
def phial(ssh_service):
    return Phial(ssh_service)


@pytest.fixture
def regex():
    return Regex()


class Regex:
    def __init__(self):
        import re
        self.re = re.compile('^$')

    def pattern(self, re):
        # type: (str) -> Regex
        self.re = re
        return self

    def matches(self, text, strip=True):
        __tracebackhide__ = True
        if strip:
            testable = text.strip()
        else:
            testable = text
        if self.re.match(testable) is None:
            pytest.fail("regex of '{regex}' do not match given text {text}"
                        .format(text=repr(testable), regex=self.re.pattern))



from puppeter import container
from puppeter.domain.model import Collection3xInstaller
from puppeter.domain.model.osfacts import OperatingSystemCodename
from puppeter.persistence.gateway.installer.debian import DebianPC3xConfigurer
from tests.domain.mock_facter import MockFacter


def test_pc3x():
    # given
    installer = Collection3xInstaller()
    installer.read_raw_options({'mode': 'Server'})
    configurer = DebianPC3xConfigurer(installer=installer)

    # when
    commands = configurer.produce_commands()
    commands = list(map(lambda cmd: cmd.strip(), commands))

    # then
    assert 'apt-get update -m' in commands
    assert 'apt-get install -y wget' in commands
    assert "wget 'https://apt.puppetlabs.com/puppetlabs-release-trusty.deb'" in commands
    assert 'apt-get install -y puppet' in commands
    assert 'puppet resource package puppetmaster ensure=installed' in commands


def setup_function():
    MockFacter.set_fact(OperatingSystemCodename, OperatingSystemCodename('trusty'))
    container.initialize()


def teardown_function():
    MockFacter.reset_facts()



from puppeter import container
from puppeter.domain.model.configurer import Configurer
from puppeter.persistence.gateway.installer.linux import RubyGemConfigurer

container.bind(Configurer, RubyGemConfigurer)



from puppeter.domain.facter import Facter

TEST_KEY = 'some-key'


def test_facter_get_unknown():
    value = Facter.get(TEST_KEY)

    assert value is None


def test_invalid_resolver():
    # noinspection PyUnresolvedReferences,PyProtectedMember
    resolvers = Facter._Facter__resolvers
    before_count = len(resolvers)
    try:
        Facter.set(TEST_KEY, lambda: None)
        value = Facter.get(TEST_KEY)

        assert value is None
        assert len(resolvers) == before_count + 1
    finally:
        resolvers.pop(TEST_KEY)
        assert len(resolvers) == before_count



from abc import ABCMeta

from six import with_metaclass

from puppeter import container
from puppeter.container import Named
from puppeter.domain.gateway.installer import InstallerGateway
from puppeter.domain.model.configurer import Configurer


@Named('gem')
class RubyGemConfigurer(Configurer):
    def __init__(self, installer):
        self.installer = installer

    def produce_commands(self):
        raise NotImplementedError('Not yet implemented!')


class LinuxInstallerGateway(with_metaclass(ABCMeta, InstallerGateway)):
    def _puppet_services(self, installer):
        return tuple(container.get_all_with_name_starting_with(Configurer,
                                                               'puppet.service',
                                                               installer=installer))

    def _puppet_cert_issue(self, installer):
        return tuple(container.get_all_with_name_starting_with(Configurer,
                                                               'puppet.cert.issue',
                                                               installer=installer))

    def _puppet_conf(self, installer):
        return tuple(container.get_all_with_name_starting_with(Configurer,
                                                               'puppet.conf',
                                                               installer=installer))



"""Puppeter - an automatic puppet installer
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
import codecs
import os
import re
import fnmatch

here = os.path.abspath(os.path.dirname(__file__))

package_data = []
for root, dirnames, filenames in os.walk('puppeter'):
    fil = fnmatch.filter(filenames, '*.sh')
    fil.extend(fnmatch.filter(filenames, '*.pp'))
    fil.extend(fnmatch.filter(filenames, '*.pyi'))
    for filename in fil:
        package_data.append(os.path.join(root, filename).replace('puppeter/', ''))


def read(*parts):
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    return codecs.open(os.path.join(here, *parts), 'r').read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name='puppeter',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=find_version('puppeter', '__init__.py'),

    description='Puppeter - an automatic puppet installer',
    # Get the long description from the README file
    long_description=read('README.rst'),

    # The project's main homepage.
    url='https://github.com/coi-gov-pl/puppeter',

    # Author details
    author='Center for Information Technology',
    author_email='opensource@coi.gov.pl',

    # Choose your license
    license='Apache-2.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Environment :: Console',
        'Environment :: Console :: Curses',
        'Operating System :: POSIX :: Linux',
        'Topic :: System :: Installation/Setup',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    # What does your project relate to?
    keywords='puppet installation devops unattended',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests', 'integration_tests']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'six',
        'ruamel.yaml<0.15',
        'distro',
        'enum34',
        'typing'
    ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['setuptools>=18.5', 'check-manifest'],
        'test': ['pytest', 'pytest-cov', 'python-coveralls', 'colorama', 'paramiko'],
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
        'puppeter': package_data,
    },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    data_files=[
      # ('puppeter', package_data)
    ],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'puppeter=puppeter.main:main',
        ],
    },
)



from __future__ import absolute_import
import os
import re
import pkg_resources

from collections import OrderedDict
from string import Template
from tempfile import NamedTemporaryFile
from typing import Type, Any
from six import iteritems

from puppeter.domain.model.configurer import Configurer, ScriptFormat, CommandsCollector


class AtTemplate(Template):
    delimiter = '@'
    flags = re.MULTILINE


class CommandsCollectorImpl(CommandsCollector):
    __collected = 0

    def __init__(self, configurer):
        self.__parts = OrderedDict()  # type: OrderedDict
        self.__configurer = configurer  # type: Configurer

    def collect_from_template(self, description, template, mapping, format=ScriptFormat.BASH):
        script = self.__template(template) \
            .substitute(mapping)
        self.__parts[description] = dict(script=script, format=format)
        return self

    def collect_from_file(self, description, path, format=ScriptFormat.BASH):
        script = self.__resource_string(path)
        self.__parts[description] = dict(script=script, format=format)
        return self

    def lines(self):
        lines = []
        for (description, script) in iteritems(self.__parts):
            CommandsCollectorImpl.__collected += 1
            lines.append('# Part %d: %s' % (CommandsCollectorImpl.__collected, description))
            lines += self.__get_linesof_script(script)
        return lines

    def __template(self, path):
        # type: (str) -> AtTemplate
        return AtTemplate(self.__resource_string(path))

    def __resource_string(self, path):
        # noinspection PyTypeChecker
        return self.__load_resource(pkg_resources.resource_string, self.__configurer, path) \
            .decode("utf-8")

    @staticmethod
    def __load_resource(loader, obj, path):
        mod = CommandsCollectorImpl.__moduleof(obj)
        return loader(mod, path)

    @staticmethod
    def __moduleof(cls):
        # type: (Type[Any]) -> str
        try:
            while True:
                cls = cls.original_cls()
        except AttributeError:
            pass
        return cls.__module__

    def __get_linesof_script(self, script):
        format = script['format']  # type: ScriptFormat
        shellscript = script['script']  # type: str
        if format is ScriptFormat.PUPPET:
            tpl = self.__load_resource(pkg_resources.resource_string, self, 'puppet-apply.sh') \
                .decode("utf-8")
            f = NamedTemporaryFile(delete=False)
            tmpfilename = f.name
            f.close()
            os.unlink(f.name)
            tpl = AtTemplate(tpl)
            shellscript = tpl.substitute(dict(tmpfilename=tmpfilename, pp=shellscript))

        return shellscript.split("\n")



from typing import Any, Callable, Dict, TypeVar, Optional, IO

import puppeter
from puppeter import container
from puppeter.domain.model.csr import CsrAttributesConfiguration
from puppeter.domain.model.fqdn import FqdnConfiguration
from puppeter.domain.model.installer import Installer
from puppeter.domain.model.answers import Answers
from puppeter.domain.gateway.answers import AnswersGateway
from puppeter.container import Named
import ruamel.yaml

T = TypeVar('T')


@Named('yaml')
class YamlAnswersGateway(AnswersGateway):
    def write_answers_to_file(self, answers, target_file):
        # type: (Answers, IO) -> None
        raw_answers = {}
        if answers.installer() is not None:
            raw_answers['installer'] = answers.installer().raw_options()
        if answers.fqdn_configuration() is not None:
            raw_answers['fqdn'] = answers.fqdn_configuration().raw_options()
        if answers.csrattrs_configuration() is not None:
            raw_answers['csr-attributes'] = answers.csrattrs_configuration().raw_options()
        yaml = ruamel.yaml.dump(raw_answers, Dumper=ruamel.yaml.RoundTripDumper)
        target_file.write(yaml)

    def read_answers_from_file(self, target_file):
        code = ruamel.yaml.load(target_file.read(), ruamel.yaml.RoundTripLoader)
        log = puppeter.get_logger(YamlAnswersGateway)
        log.debug("Answers loaded from file: %s", code)
        installer = self.__process_data(code,
                                        'installer',
                                        lambda val: self.__load_installer(val))
        fqdn = self.__process_data(code,
                                   'fqdn',
                                   lambda val: FqdnConfiguration(val))
        csr_attrs = self.__process_data(code,
                                        'csr-attributes',
                                        lambda val: self.__load_csr(val))
        return Answers(
            installer=installer,
            fqdn_configuration=fqdn,
            csrattrs_configuration=csr_attrs
        )

    @staticmethod
    def __load_csr(options):
        conf = CsrAttributesConfiguration()
        conf.read_raw_options(options)
        return conf

    @staticmethod
    def __load_installer(options):
        bean_name = options['type']
        installer = container.get_named(Installer, bean_name)
        installer.read_raw_options(options)
        return installer

    @staticmethod
    def __process_data(data, key, func):
        # type: (Dict[str,Any],str,Callable[[Any],T]) -> Optional[T]
        try:
            value = data[key]
            return func(value)
        except KeyError:
            return None



from typing import Sequence, List

import puppeter
from puppeter import container
from puppeter.domain.facter import Facter
from puppeter.domain.gateway.answers import AnswersProcessor
from puppeter.domain.gateway.csr import CsrAttributesSetterGateway
from puppeter.domain.gateway.fqdn import FqdnSetterGateway
from puppeter.domain.gateway.installer import InstallerGateway
from puppeter.domain.gateway.script import ScriptPostProcessor, ScriptLibrariesConfigurer
from puppeter.domain.model import osfacts
from puppeter.domain.model.answers import Answers
from puppeter.domain.model.configurer import Configurer
from puppeter.presentation.app import Options


class AnswersProcessorImpl(AnswersProcessor):

    def __init__(self, options):
        self.options = options  # type: Options
        self.__log = puppeter.get_logger(AnswersProcessorImpl)

    def process(self, answers):
        configurers = []
        configurers.extend(self.__attach_libs())
        configurers.extend(self.__perform_installation(answers))
        configurers.extend(self.__setup_fqdn(answers))
        configurers.extend(self.__setup_csr(answers))
        commands = self.__collect_commands(configurers)
        for line in commands:
            if self.options.execute():
                self.__log.warning('EXECUTING: %s', line)
            else:
                print(line)

    @staticmethod
    def __perform_installation(answers):
        # type: (Answers) -> Sequence[Configurer]
        configurers = []
        if answers.installer() is not None:
            osfamily = Facter.get(osfacts.OsFamily)
            gateway = container.get_named(InstallerGateway, osfamily.name.lower())
            configurers.extend(
                gateway.configurers(answers.installer())
            )
        return configurers

    @staticmethod
    def __setup_fqdn(answers):
        # type: (Answers) -> Sequence[Configurer]
        configurers = []
        if answers.fqdn_configuration() is not None:
            gateway = container.get(FqdnSetterGateway)
            configurers.extend(
                gateway.process_fully_qualified_domain_name(answers.fqdn_configuration())
            )
        return configurers

    @classmethod
    def __collect_commands(cls, configurers):
        # type: (List[Configurer]) -> Sequence[str]
        sorted_cfg = sorted(configurers, key=AnswersProcessorImpl.__sort_by_order)
        commands = []
        for configurer in sorted_cfg:
            commands.extend(configurer.produce_commands())
        return cls.__postprocess(commands)

    @staticmethod
    def __sort_by_order(configurer):
        # type: (Configurer) -> int
        try:
            # noinspection PyUnresolvedReferences
            return configurer.order()
        except AttributeError:
            return 500

    @classmethod
    def __postprocess(cls, commands):
        # type: (Sequence[str]) -> Sequence[str]
        postprocessor = container.get(ScriptPostProcessor)
        return postprocessor.postprocess(commands)

    @staticmethod
    def __attach_libs():
        # type: () -> Sequence[Configurer]
        return [
            container.get(ScriptLibrariesConfigurer)
        ]

    @staticmethod
    def __setup_csr(answers):
        # type: (Answers) -> Sequence[Configurer]
        if answers.csrattrs_configuration() is not None:
            csrsetter = container.get(CsrAttributesSetterGateway)
            return csrsetter.save_csr_attributes(
                answers.csrattrs_configuration(),
                answers.installer().mode()
            )
        return []



import re

from puppeter.domain.gateway.script import ScriptPostProcessor, ScriptLibrariesConfigurer
from puppeter.domain.model.ordered import Order


class BashScriptPostProcessor(ScriptPostProcessor):

    SHEBANG_REGEX = re.compile('^#!(.*)\n', re.MULTILINE)

    def postprocess(self, commands):
        ln = "\n"
        header = ['#!/usr/bin/env bash', 'set -eux -o pipefail', '']
        stripped = self.__strip_shebang(ln.join(commands)).split(ln)
        return header + stripped

    @staticmethod
    def __strip_shebang(script):
        return re.sub(BashScriptPostProcessor.SHEBANG_REGEX, '', script)


@Order(-1000)
class BashScriptLibrariesConfigurer(ScriptLibrariesConfigurer):
    def produce_commands(self):
        return self._collector()\
            .collect_from_file('Bash script libraries', 'bash-libraries.sh')\
            .lines()



from puppeter.domain.model.fqdn import FqdnConfiguration


def test_fqdn_read():
    # given
    fqdn = FqdnConfiguration()

    # when
    fqdn.read_raw_options(options='app01.acme.internal')

    # then
    assert fqdn.fqdn() == 'app01.acme.internal'
    assert fqdn.hostname() == 'app01'
    assert fqdn.has_domain()
    assert fqdn.domain() == 'acme.internal'


def test_fqdn_simple():
    # given
    fqdn = FqdnConfiguration('simple')

    # then
    assert fqdn.fqdn() == 'simple'
    assert fqdn.hostname() == 'simple'
    assert fqdn.domain() is None
    assert fqdn.has_domain() is False



import os

from puppeter import container
from puppeter.container import Named, Container
from puppeter.domain.model.answers import Answers
from puppeter.presentation import App
from puppeter.presentation.cmdparser import CommandLineParser

HERE = os.path.abspath(__file__)


@Named('interactive')
class InteractiveTestApp(App):
    def _collect_answers(self):
        return Answers()

    def run(self):
        return 'runned interactivly'


@Named('unattended')
class UnattendedTestApp(App):
    def _collect_answers(self):
        return Answers()

    def run(self):
        return 'runned unattended'


class TestAppCreation:
    old_container = None

    @classmethod
    def setup_class(cls):
        cls.old_container = container.app_container
        container.app_container = Container()
        container.bind(App, InteractiveTestApp)
        container.bind(App, UnattendedTestApp)

    @classmethod
    def teardown_class(cls):
        container.app_container = cls.old_container

    def test_proper_app_creation(self):
        # given
        parser = CommandLineParser(['puppeter'])
        # when
        app = parser.parse()
        result = app.run()
        # then
        assert app is not None
        assert result == 'runned interactivly'

    def test_proper_app_creation_unattended(self):
        # given
        parser = CommandLineParser(['puppeter', '--answers', HERE])
        # when
        app = parser.parse()
        result = app.run()
        # then
        assert app is not None
        assert result == 'runned unattended'
